import os
import warnings
from typing import Dict, List

from sqlalchemy import create_engine, select
from sqlalchemy.orm import sessionmaker, session
from sqlalchemy_utils import database_exists, create_database

from timelink.api import database
from timelink import mhk
from timelink.app.models.user import User, UserProperty, Base
from timelink.app.models.project import ProjectAccess, Project
from timelink.app.schemas.user import UserPropertySchema, UserSchema
from timelink.app.schemas import UserProjectSchema
from timelink.app.services.auth import FiefUserInfo


class UserDatabase:
    def __init__(
        self,
        db_name: str = " timelink_users.sqlite",
        db_type: str = "sqlite",
        db_url=None,
        db_user=None,
        db_pwd=None,
        db_path=None,
        postgres_image=None,
        postgres_version=None,
        initial_users=None,
        stop_duplicates=True,
        **connect_args,
    ):
        """Create a new UserDatabase object.

        Args:
            db_name: the name of the database
            db_type: the type of the database (sqlite, postgres)
            db_url: the url of the database
            db_user: the user to access the database
            db_pwd: the password to access the database
            db_path: the path to the database file
            postgres_image: the docker image of the postgres server
            postgres_version: the version of the postgres server
            stop_duplicates: if True, stop duplicates in the database

        Usage:
            ``python
            db = UserDatabase(db_name='test-users')
            with db.session() as session:
                user = User(name='Test User', fullname='Test User Fullname', email='xpto@xpto.com', nickname='Test Nickname')
                session.add(user)
                session.commit()

            # or manage the session in UserDatabase
            db.start_session()  # returns a session

            db.add_user(user)  # no need to specify the session
            db.commit()   # or db.rollback()
            db.close_session()

            # if you want to use the same session
            session = db.current_session``

        """
        self.db_name = None
        self.db_url = None
        self.db_user = None
        self.db_pwd = None
        self.db_path = None
        self.db_type = db_type
        self.db_container = None
        self.engine = None
        self.session = None
        self.current_session = None
        self.metadata = None
        self.cache: Dict[str, UserSchema] = {}

        if db_name is None:
            raise ValueError("db_name cannot be None")
        self.db_name = db_name

        if self.db_type is None:
            self.db_type = "sqlite"

        # if we received a url, use it to connect
        if db_url is not None:
            self.db_url = db_url
        # if not, generate a url from other parameters
        else:
            if db_type == "sqlite":
                if db_name == ":memory:":
                    self.db_url = "sqlite:///:memory:"
                else:
                    if db_path is None:
                        db_path = os.getcwd()
                        warnings.warn(
                            "No path for the database was specified. Using the current working directory.",
                            stacklevel=2,
                        )
                    if connect_args is None:
                        connect_args = {"check_same_thread": False}
                    db_path = os.path.abspath(db_path)
                    os.makedirs(db_path, exist_ok=True)
                    self.db_url = f"sqlite:///{db_path}/{self.db_name}"
            elif db_type == "postgres":
                if db_pwd is None:
                    self.db_pwd = database.get_db_password()
                else:
                    self.db_pwd = db_pwd
                if db_user is None:
                    self.db_user = "postgres"
                else:
                    self.db_user = db_user
                if database.is_postgres_running():
                    self.db_container = database.get_postgres_container()
                    # if it it is running, we need the password
                    container_vars = self.db_container.attrs["Config"]["Env"]
                    pwd = [var for var in container_vars if "POSTGRES_PASSWORD" in var][
                        0
                    ]
                    self.db_pwd = pwd.split("=")[1]
                    usr = [var for var in container_vars if "POSTGRES_USER" in var][0]
                    self.db_user = usr.split("=")[1]
                else:
                    self.db_container = database.start_postgres_server(
                        self.db_name,
                        self.db_user,
                        self.db_pwd,
                        image=postgres_image,
                        version=postgres_version,
                    )
                self.db_url = (
                    f"postgresql://{self.db_user}:"
                    f"{self.db_pwd}@127.0.0.1/{self.db_name}"
                )
                self.db_container = database.start_postgres_server(
                    self.db_name, self.db_user, self.db_pwd
                )
                self.db_pwd = database.get_postgres_container_pwd()
            elif db_type == "mysql":
                self.db_url = f"mysql://{db_user}:{db_pwd}@localhost/{db_name}"
                if db_pwd is None:
                    try:
                        self.db_pwd = mhk.utilities.get_mhk_db_pwd()
                    except TypeError:
                        self.db_pwd = None
                if db_pwd is None:
                    self.db_pwd = database.random_password()
                # TODO Start a mysql server in docker
            else:
                raise ValueError(f"Unknown database type: {db_type}")

        self.engine = create_engine(self.db_url, connect_args=connect_args)
        if not database_exists(self.engine.url):
            create_database(self.engine.url)
        self.session = sessionmaker(autocommit=False, autoflush=False, bind=self.engine)
        self.current_session = None
        self.metadata = Base.metadata
        Base.metadata.create_all(self.engine)
        if initial_users is not None:
            with self.session() as session:
                for user in initial_users:
                    user_exists = self.get_user_by_name(user.name, session=session)
                    if user_exists is None:
                        self.add_user(user, session=session)
                session.commit()

    def __enter__(self):
        self.current_session = self.session()
        return self.current_session

    def __exit__(self, exc_type, exc_value, traceback):
        self.current_session.close()

    def drop_db(self):
        """Drop the database"""
        self.metadata.drop_all(self.engine)

    def commit(self):
        """Commit the current session"""
        self.current_session.commit()

    def rollback(self):
        """Rollback the current session"""
        self.current_session.rollback()

    def add_user(self, user: User, session: session = None):
        """Add a user to the database

        Args:
            user: the user to add

        """
        if session is None:
            if self.current_session is None:
                raise ValueError(
                    "No session available."
                    "Either: pass a session=session or  "
                    "do db.start_session() before and db.close_session() or "
                    "use context manager with db:"
                )
            else:
                session = self.current_session

        if self.get_user_by_email(user.email, session=session) is not None:
            raise ValueError("An user with that email already exists.")

        session.add(user)

    def on_board_user(self, user: FiefUserInfo, permissions: str = None, session=None):
        """On board an user registered on Fief

        Gets email, name and fields from fief and adds it to the database.

        Adds permissions if provided. Permissions should be fetched
        from the fief access_token and passed as a comma separeted string.

        The access_token is stored in a cookie and can be validated with
        fief.validate_access_token

        """

        user_email = user["email"]
        user_in_db = self.get_user_by_email(user_email, session=session)
        if user_in_db is None:
            user_fields = user.get("fields", {})
            user_name = user_fields.get("user_name", "")
            user_nickname = user_fields.get("user_nickname", None)
            if user_nickname is None:
                user_nickname = user_email.split("@")[0]
            user = User(name=user_name, nickname=user_nickname, email=user_email)
            self.add_user(user, session=session)
            session.commit()
        for field_name, field_value in user["fields"].items():
            self.set_user_property(
                user_in_db.id, field_name, field_value, session=session
            )
        session.commit()
        # todo: think about this
        for permission in permissions:
            self.set_user_property(
                user_in_db.id, f"permission:{permission}", "yes", session=session
            )
        session.commit()

        return UserSchema.model_validate(user_in_db)

    def set_user_property(
        self, user_id: int, property_name: str, property_value: str, session=None
    ):
        """Add a user property to the database

        Args:
            user_id: the id of the user
            property_name: the name of the property
            property_value: the value of the property

        """
        if session is None:
            if self.current_session is None:
                raise ValueError(
                    "No session available."
                    "Either pass a session or "
                    "with db: , with db as session"
                )
            else:
                session = self.current_session
        # check if property already exists

        user_property = session.scalars(
            select(UserProperty).filter(
                UserProperty.user_id == user_id, UserProperty.name == property_name
            )
        ).first()
        if user_property is not None:
            user_property.value = property_value
            session.merge(user_property)
        else:
            user_property = UserProperty(
                name=property_name, value=property_value, user_id=user_id
            )  # pylint: disable=no-member
            session.add(user_property)

    def get_user(self, user_id: int, session=None) -> User:
        """Get a user from the database

        Args:
            user_id: the id of the user to get

        Returns:
            the user with the given id

        """
        if session is None:
            if self.current_session is None:
                raise ValueError(
                    "No session available."
                    "Either pass a session or "
                    "with session=db:"
                )
            else:
                session = self.current_session
        user: User = session.get(User, user_id)
        if user is None:
            return None
        else:
            return user

    def get_user_by_name(self, name: str, session=None) -> User:
        """Get a user from the database

        Args:
            name: the name of the user to get
            session: a database session

        Returns:
            the user with the given name

        """
        if session is None:
            if self.current_session is None:
                raise ValueError(
                    "No session available."
                    "Either pass a session or "
                    "with session=db:"
                )
            else:
                session = self.current_session

        user: User = session.scalars(select(User).filter(User.name == name)).first()
        if user is None:
            return None
        else:
            return user

    def get_user_by_email(self, email, session=None):
        """
        Retrieve a user from the database using their email address.

        Args:
            email (str): The email address of the user to retrieve.
            session (Session, optional): The session to use for the query.
                If None, the current session will be used. Defaults to None.

        Returns:
            User: The User object that matches the given email address,
                or None if no user was found.
        """
        if session is None:
            if self.current_session is None:
                raise ValueError(
                    "No session available."
                    "Either pass a session or "
                    "with db as session:"
                )
            else:
                session = self.current_session

        user: User = session.scalars(select(User).filter(User.email == email)).first()
        if user is None:
            return None
        else:
            return user

    def get_user_by_nickname(self, nickname, session=None):
        """
        Retrieve a user from the database using their nickname.

        Args:
            nickname (str): The nickname address of the user to retrieve.
            session (Session, optional): The session to use for the query.
                If None, the current session will be used. Defaults to None.

        Returns:
            User: The User object that matches the given nickname,
                or None if no user was found.
        """
        if session is None:
            if self.current_session is None:
                raise ValueError(
                    "No session available."
                    "Either pass a session or "
                    "with db as session:"
                )
            else:
                session = self.current_session

        user: User = session.scalars(
            select(User).filter(User.nickname == nickname)
        ).first()
        if user is None:
            return None
        else:
            return user

    def update_user(self, user: User, session=None):
        """Update a user in the database

        Args:
            user: the user to update

        """
        if session is None:
            if self.current_session is None:
                raise ValueError(
                    "No session available."
                    "Either pass a session or "
                    "with db as session:"
                )
            else:
                session = self.current_session

        session.merge(user)

    def delete_user(self, user: User, session=None):
        """Delete a user from the database

        Args:
            user: the user to delete

        """
        if session is None:
            if self.current_session is None:
                raise ValueError(
                    "No session available."
                    "Either pass a session or "
                    "with db as session:"
                )
            else:
                session = self.current_session

        session.delete(user)
        session.commit()

    def get_user_properties(
        self, user_id: int, session=None
    ) -> List[UserPropertySchema]:
        """Get the properties of a user

        Args:
            user_id: the id of the user

        Returns:
            a list with the properties of the user

        """
        if session is None:
            if self.current_session is None:
                raise ValueError(
                    "No session available."
                    "Either pass a session or "
                    "with db as session:"
                )
            else:
                session = self.current_session
        result = session.scalars(
            select(UserProperty).filter(UserProperty.user_id == user_id)
        ).all()
        return result

    def get_user_property(
        self, user_id: int, property_name: str, session=None
    ) -> UserPropertySchema:
        """Get a user property

        Args:
            user_id: the id of the user
            property_name: the name of the property

        Returns:
            the property with the given name

        """
        if session is None:
            if self.current_session is None:
                raise ValueError(
                    "No session available."
                    "Either pass a session or "
                    "with db as session:"
                )
            else:
                session = self.current_session

        r = session.scalars(
            select(UserProperty).filter(
                UserProperty.user_id == user_id, UserProperty.name == property_name
            )
        ).first()
        return r

    def set_user_project_access(
        self, user_id: int, project_id: int, access_level: str, session=None
    ):
        """Set the access level of a user to a project

        Args:
            user_id: the id of the user
            project_id: the id of the project
            access_level: the access level

        """
        if session is None:
            if self.current_session is None:
                raise ValueError(
                    "No session available."
                    "Either pass a session or "
                    "with db as session:"
                )
            else:
                session = self.current_session

        project_access = session.scalars(
            select(ProjectAccess).filter(
                ProjectAccess.user_id == user_id, ProjectAccess.project_id == project_id
            )
        ).first()
        if project_access is not None:
            project_access.access_level = access_level
            session.merge(project_access)
        else:
            project_access = ProjectAccess(
                user_id=user_id, project_id=project_id, access_level=access_level
            )
            session.add(project_access)

    def get_user_project_access(
        self, user_id: int, project_id: int, session=None
    ) -> ProjectAccess:
        """Get the access level of a user to a project

        Args:
            user_id: the id of the user
            project_id: the id of the project

        Returns:
            the access level of the user to the project

        """
        if session is None:
            if self.current_session is None:
                raise ValueError(
                    "No session available."
                    "Either pass a session or "
                    "with db as session:"
                )
            else:
                session = self.current_session

        return session.scalars(
            select(ProjectAccess).filter(
                ProjectAccess.user_id == user_id, ProjectAccess.project_id == project_id
            )
        ).first()

    def get_user_projects(self, user_id: int, session=None) -> List[UserProjectSchema]:
        """Get the projects of a user

        Args:
            user_id: the id of the user

        Returns:
            a list with the projects of the user

        """
        if session is None:
            if self.current_session is None:
                raise ValueError(
                    "No session available."
                    "Either pass a session or "
                    "with db as session:"
                )
            else:
                session = self.current_session
        stmt = (
            select(
                Project.id,
                Project.name,
                Project.description,
                ProjectAccess.access_level,
            )
            .join(Project.users)
            .filter(ProjectAccess.user_id == user_id)
        )
        result = session.execute(stmt).all()

        projects = []
        for pid, name, description, user_access in result:
            user_project = UserProjectSchema(
                project_id=pid,
                user_id=user_id,
                project_name=name,
                project_description=description,
                access_level=user_access,
            )
            projects.append(user_project)

        return projects

    def get_project(self, project_id: int, session=None) -> Project:
        if session is None:
            if self.current_session is None:
                raise ValueError(
                    "No session available."
                    "Either pass a session or "
                    "with session=db:"
                )
            else:
                session = self.current_session

        return session.get(Project, project_id)

    def get_project_by_name(self, name: str, session=None) -> Project:
        if session is None:
            if self.current_session is None:
                raise ValueError(
                    "No session available."
                    "Either pass a session or "
                    "with session=db:"
                )
            else:
                session = self.current_session

        result = session.scalars(select(Project).filter(Project.name == name))
        if result is not None:
            return result.first()
        else:
            return None



from fastapi import Request

from fastui import AnyComponent, components as c
from fastui.events import GoToEvent

from timelink import version as timelink_version
from timelink.app.schemas.user import UserSchema


async def home_page(
    *components: AnyComponent,
    request: Request = None,
    title: str = None,
    user: UserSchema = None,
) -> list[AnyComponent]:
    """Timelink FastUI main page

    This function creates the basic layout of the Timelink FastUI main page.
    It adds specific componenents passed as the first argument to the page.
    Also if uses the user parameter to configure the login/logout link, and
    change the options available in the navbar.

    The current app instance is obtained from the request object

    Args:
        *components (AnyComponent): Components to add to the page
        request (Request, optional): The request. Defaults to None.
        title (str, optional): The title of the page. Defaults to None.
        user (str, optional): The user logged in. Defaults to None.
    """
    if user is None:
        user_name = "guest"
        current_project_name = "None"
    else:
        user_name = user.name
        if user_name == "":
            user_name = user.email.split("@")[0]
        if user.current_project_name is None:
            current_project_name = "None"
        else:
            current_project_name = user.current_project_name

    start_links = [
        c.Link(
            components=[c.Text(text="Explore")],
            on_click=GoToEvent(url="/explore"),
            active="startswith:/explore",
        ),
        c.Link(
            components=[c.Text(text="Sources")],
            on_click=GoToEvent(url="/sources"),
            active="startswith:/sources",
        ),
        c.Link(
            components=[c.Text(text="Info")],
            on_click=GoToEvent(url="/info"),
            active="startswith:/info",
        ),
        c.Link(
            components=[c.Text(text="Projects")],
            on_click=GoToEvent(url="/projects"),
            active="startswith:/projects",
        ),
    ]
    if user is not None and user.is_admin():
        admin_link = c.Link(
            components=[c.Text(text="Admin")],
            on_click=GoToEvent(url="/adm"),
            active="startswith:/adm",
        )
        start_links.append(admin_link)

    if user is None or user_name == "guest":
        auth_url = request.url_for("login")
        login_logout = c.Link(
            components=[c.Text(text="Login")],
            on_click=GoToEvent(url=str(auth_url)),
            active=f"startswith:{auth_url}",
        )
    else:
        logout_url = request.url_for("logout")
        login_logout = c.Link(
            components=[c.Text(text="Logout")],
            on_click=GoToEvent(url=str(logout_url)),
            active=f"startswith:{logout_url}",
        )
    end_links = [login_logout]
    page = [
        c.PageTitle(text=f"{title}" if title else "Timelink"),
        c.Navbar(
            title="Timelink",
            title_event=GoToEvent(url="/"),
            start_links=start_links,
            end_links=end_links,
        ),
        c.Page(
            components=[
                *((c.Heading(text=title, level=3),) if title else ()),
                *components,
                c.Heading(level=6, text=f"{user_name} ({current_project_name})"),
            ],
        ),
        c.Footer(
            extra_text="timelink-py version: " + timelink_version,
            links=[
                c.Link(
                    components=[c.Text(text="Source code")],
                    on_click=GoToEvent(url="https://github.com/time-link/timelink-py"),
                ),
                c.Link(
                    components=[c.Text(text="PyPI")],
                    on_click=GoToEvent(url="https://pypi.org/project/timelink/"),
                ),
                c.Link(
                    components=[c.Text(text="Docs")],
                    on_click=GoToEvent(
                        url="https://joaquimrcarvalho.github.io/timelink/"
                    ),
                ),
                c.Link(
                    components=[c.Text(text="API documentation")],
                    on_click=GoToEvent(
                        url=f"{request.url.scheme}://{request.url.netloc}/docs"
                    ),
                ),
                c.Link(
                    components=[c.Text(text="Redoc documentation")],
                    on_click=GoToEvent(
                        url=f"{request.url.scheme}://{request.url.netloc}/redoc"
                    ),
                ),
            ],
        ),
    ]
    return page



""" Authentication services """
import os
import uuid
from typing import Optional
from typing_extensions import Dict
from fastapi import Request, Response, HTTPException, status
from fastapi.security import APIKeyCookie

# see https://docs.fief.dev/integrate/python/fastapi/#web-application-example
from fief_client import FiefAsync, FiefUserInfo
from fief_client.integrations.fastapi import FiefAuth


class CustomFiefAuth(FiefAuth):
    """
    CustomFiefAuth is a subclass of FiefAuth that allows for customization of user authentication.

    This class is designed to be used in scenarios where the base FiefAuth class does not provide
    the desired behavior. Specifically, it allows for customization of the response when a user
    is not authenticated or does not have the required scope.

    Attributes:
        client (FiefAsync): An instance of the FiefAsync class for making asynchronous requests.

    See https://fief-dev.github.io/fief-python/fief_client/integrations/fastapi.html#FiefAuth.__init__ for more details.
    """
    client: FiefAsync

    async def get_unauthorized_response(self,
                                        request: Request,
                                        response: Response):
        redirect_uri = request.url_for("auth_callback")
        auth_url = await self.client.auth_url(redirect_uri, scope=["openid"])
        # Redirect to the Fief auth URL
        raise HTTPException(
            status_code=status.HTTP_307_TEMPORARY_REDIRECT,
            headers={"Location": str(auth_url)}
        )


class MemoryUserInfoCache:
    def __init__(self) -> None:
        self.storage: Dict[uuid.UUID, FiefUserInfo] = {}

    async def get(self, user_id: uuid.UUID) -> Optional[FiefUserInfo]:
        return self.storage.get(user_id)

    async def set(self, user_id: uuid.UUID, userinfo: FiefUserInfo) -> None:
        self.storage[user_id] = userinfo


memory_userinfo_cache = MemoryUserInfoCache()


async def get_memory_userinfo_cache() -> MemoryUserInfoCache:
    return memory_userinfo_cache


fief_url = os.getenv("FIEF_URL")
fief_key = os.getenv("FIEF_KEY")
fief_secret = os.getenv("FIEF_SECRET")
fief = FiefAsync(fief_url, fief_key, fief_secret)

SESSION_COOKIE_NAME = "user_session"
scheme = APIKeyCookie(name=SESSION_COOKIE_NAME, auto_error=False)
auth = CustomFiefAuth(
    fief,
    scheme,
    get_userinfo_cache=get_memory_userinfo_cache,
)



from timelink.kleio.groups.kgroup import KGroup


class KKleio(KGroup):
    """KKleio(structure,prefix=,obs=,translations=,translator=)

    Kleio notation document. Represents a file in Kleio notation.

    Elements:
        structure: The path to a Kleio structure file (default gacto2.str)
        prefix: Prefix to be added to all ids generated from this file
        translations: number of times this file was translated
        translator: name of the translator to be used (currently not used)
        obs: observations

    TODO: Could keep a list of the name of the groups included, at all levels
    """

    _name = "kleio"
    _position = ["structure"]
    _also = ["prefix", "translations", "translator", "obs"]
    _part = ["source", "aregister"]
    _pom_class_id: str = "entity"

    def __init__(self, *args, **kwargs):
        KGroup._global_line = 1
        KGroup._global_sequence = 1
        super().__init__(*args, **kwargs)



"""
Utilities for interaction with and MHK installation.

MHK is the name of Java Webapp that preceded Timelink.

(c) Joaquim Carvalho, 2021. MIT Licence.

"""

import warnings
from collections import namedtuple
from configparser import ConfigParser
from pathlib import Path
from typing import Type, Union

from sqlalchemy import create_engine, text


def get_env_as_dict(filename: str) -> dict:
    """
    Reads environment variables from a file and return a SafeConfig object

    Rationale: SafeConfig parser requires
    :param filename: Name of a file
    :return:
    """
    conf = "[env]\n"
    with open(filename, "r") as f:
        env_file = f.read()  # read from file
        conf = conf + env_file

    env = ConfigParser()
    # preserve case in env vars
    # see https://stackoverflow.com/questions/1611799/preserve-case-in-configparser # noqa: E501
    env.optionxform = str

    env.read_string(conf)

    return remove_quotes(dict(env["env"].items()))


def remove_quotes(original):
    """
    Removes quotes from quoted values in env variables
    See https://stackoverflow.com/a/50772706
    :param original: dict with env variables
    :return: copy of the dict with the values unquoted
    """
    d = original.copy()
    for key, value in d.items():
        if isinstance(value, str):
            s = d[key]
            if s.startswith(('"', "'")):
                s = s[1:]
            if s.endswith(('"', "'")):
                s = s[:-1]
            d[key] = s
            # print(f"string found: {s}")
        if isinstance(value, dict):
            d[key] = remove_quotes(value)
    #
    return d


def get_mhk_env() -> Type[Union[str, None]]:
    """
    Get the MHK environment variables from ~/.mhk

    The .mhk file contains the path for the mhk-home of the
    current user. It is created by the MHK install process.
    """
    if is_mhk_installed():
        home_dir = str(Path.home())
        if home_dir is None:
            warnings.warn("Could not get a home directory", stacklevel=2)
            return None
        else:
            env = get_env_as_dict(home_dir + "/.mhk")
            if env is None:
                warnings.warn("Could not read .mhk env from user home", stacklevel=2)
            return env
    else:
        warnings.warn("MHK is not installed", stacklevel=2)
        return None


def get_mhk_app_env() -> Type[Union[str, None]]:
    """
    Get the mhk app environment variables from mhk-home/app/.env

    The .env filed contains information used by docker to create the
    MHK application and some running options, like header color and
    docker image tags used for updates.

    The .env file is created by the MHK installation process.
    Its contents can be changed my mhk manager commands.
    """
    mhk_env = get_mhk_env()
    if mhk_env is not None:
        mhk_home_dir = mhk_env["HOST_MHK_HOME"]
        app_env = get_env_as_dict(mhk_home_dir + "/app/.env")
        return app_env
    else:
        warnings.warn("Could not get MHK env variables", stacklevel=2)
        return None


def get_db_pwd() -> str:
    """
    Get the password of the database from the MHK environment
    """
    app_env = get_mhk_app_env()
    if app_env:
        pwd = app_env["MYSQL_ROOT_PASSWORD"]
        if pwd is None:
            raise TypeError("Could not find MHK database password." "Is MHK installed?")
        else:
            return pwd
    else:
        raise TypeError(
            "Could not find MHK app information." "Is MHK installed?"
        )  # noqa: E501


def get_connection_string(db: str, host="localhost", port="3307") -> str:
    pwd = get_db_pwd()
    conn_string = f"mysql+mysqlconnector://root:{pwd}@{host}:{port}/{db}"
    return conn_string


def get_engine(db: str, host="localhost", port="3307", echo=False):
    cs = get_connection_string(db, host, port)
    return create_engine(cs, echo=echo, future=True)


def get_dbnames():
    """
    Get the names of MHK databases in MySQL

    A search is made in the MySQL server running in the local host port 3307
    """
    pwd = get_mhk_db_pwd()
    conn_string = "mysql+mysqlconnector://root:{p}@localhost:3307/mysql".format(p=pwd)
    mysql = create_engine(conn_string, echo=False, future=True)
    with mysql.connect() as conn:
        databases = conn.execute(
            text(
                "SELECT table_schema FROM information_schema.tables"
                "       WHERE  table_name = 'entities'"
            )
        )
        result = [db[0] for db in databases]
    return result


def get_mhk_db_pwd():
    """
    Get the password of the database from the MHK environment
    """
    app_env = get_mhk_app_env()
    if app_env:
        pwd = app_env["MYSQL_ROOT_PASSWORD"]
        if pwd is None:
            raise TypeError("Could not find MHK database password." "Is MHK installed?")
        else:
            return pwd
    else:
        raise TypeError("Could not find MHK app information." "Is MHK installed?")


def get_mhk_info():
    mhk_env = get_mhk_env()
    user_home = mhk_env["HOST_MHK_USER_HOME"]
    mhk_home = mhk_env["HOST_MHK_HOME"]
    with open(mhk_home + "/app/manager_version", "r") as file:
        mv = file.read().replace("\n", "")
    with open(mhk_home + "/.mhk-home", "r") as file:
        mhk_home_update = file.read().replace("\n", "")
    with open(mhk_home + "/.mhk-home-manager-init", "r") as file:
        mhk_home_init = file.read().replace("\n", "")
    mhk_app_env = get_mhk_app_env()
    mhk_host = mhk_app_env.get("MHK_HOST", "localhost")

    MHKInfo = namedtuple(
        "MHKInfo",
        [
            "mhk_app_env",
            "mhk_home",
            "mhk_home_init",
            "mhk_home_update",
            "mhk_host",
            "mhk_version",
            "user_home",
        ],
    )
    mhk_info = MHKInfo(
        mhk_app_env, mhk_home, mhk_home_init, mhk_home_update, mhk_host, mv, user_home
    )
    return mhk_info


def is_mhk_installed() -> bool:
    """Returns true if a MHK instalation is found

    Checks the existence of ~/.mhk
    """
    return Path(Path.home(), ".mhk").is_file()



""" Implementation of Acts in POM model
(c) Joaquim Carvalho 2021.
MIT License, no warranties.
"""

# pylint: disable=import-error
from sqlalchemy import Column, String, ForeignKey

from timelink.mhk.models.entity import Entity
from timelink.kleio.utilities import quote_long_text, kleio_escape


class Act(Entity):
    """Represents an Act, i.e. a record of some event in a historical document.

    Examples of acts are: baptisms, marriages, purchase deeds, wills,
    court records.
    """

    __tablename__ = "acts"

    #: str:  a string uniquely identifying the act
    id = Column(String, ForeignKey("entities.id"), primary_key=True)
    #: str: the type of the act
    the_type = Column(String(32))
    #: str: the date of the act in Kleio format AAAAMMDD
    the_date = Column(String, index=True)
    #: str: location of the act, eg church, notary office
    loc = Column(String)
    ref = Column(String)  #: str: archival reference
    obs = Column(String)  #: any observations or comments.

    __mapper_args__ = {"polymorphic_identity": "act"}

    def __repr__(self):
        """String can be used to instantiate an Act"""
        sr = super().__repr__()
        return (
            f"Act(id={sr}, "
            f'the_type="{kleio_escape(self.the_type)}", '
            f'the_date="{self.the_date}", '
            f'local="{kleio_escape(self.loc)}", '
            f'ref="{kleio_escape(self.ref)}", '
            f'obs="{quote_long_text(self.obs)}"'
            f")"
        )

    def __str__(self):
        r = (
            f"{self.groupname}${self.id}"
            f"/{self.the_date}"
            f"/type={kleio_escape(self.the_type)}"
            f"/ref={quote_long_text(self.ref)}"
            f"/loc={quote_long_text(self.loc)}"
        )
        if self.obs is not None:
            r = f"{r}/obs={quote_long_text(self.obs)}"
        return r



# noqa B008
"""
FastAPI routers for the web interface of Timelink.

This is the main switch board of the FastUI web interface.

All the top level routers are defined here, and also
the subrouters, like the /auth router, that handles
the login and logout pages.

To add future pages:

1. Write a new function in a separate file that produces the
   components for the new tab: see :mod:`webapp_info` for
   inspiration.
2. Here import the page and add a new router that calls
   the new function to collect the components and call
   :func:`home_page` to wrap the components in a page.
3. If the new page handles subpaths, like /auth/login,
   it is better to create a new router in the new page
   and include it here, like the /auth router is included.
"""

import uuid

from typing import Annotated, Optional

from fastapi import APIRouter, Depends, Form, Request, Response, Query
from fastapi.responses import RedirectResponse

from fastui import FastUI, AnyComponent, components as c
from fastui.events import GoToEvent

# from timelink.app.backend import settings
from timelink.api.database import TimelinkDatabase
from timelink.app.backend.timelink_webapp import TimelinkWebApp
from timelink.app.models.project import Project, ProjectAccess
from timelink.app.schemas.user import UserSchema
from timelink.app.services.auth import fief, auth, FiefUserInfo
from timelink.app.services.auth import (
    get_memory_userinfo_cache,
    MemoryUserInfoCache,
    SESSION_COOKIE_NAME,
)
from .home_page import home_page
from .webapp_info import webapp_info
from .projects_page import projects_info

# deprecated we dont handle logins anymore
from timelink.app.dependencies import get_current_user

# --------------

router = APIRouter(tags=["fastui"], responses={404: {"description": "Not found"}})

# deprecated we dont handle logins anymore see fiefUserDep
UserDep = Annotated[UserSchema, Depends(get_current_user)]


@router.get("/login", name="login")
async def login(
    request: Request, user: FiefUserInfo = Depends(auth.current_user(optional=False))):  # noqa B008
    """We require the user to be logged in to access this page

    So the Dependency will forward the user to the fief login page
    """
    RedirectResponse(request.url_for("/"))


@router.get("/auth-callback", name="auth_callback")
async def auth_callback(
    request: Request,
    response: Response,
    code: str = Query(...),  # noqa B008
    memory_userinfo_cache: MemoryUserInfoCache = Depends(get_memory_userinfo_cache),  # noqa B008
):
    """Callback for the Fief authentication

    From the fief docs:

    "We implement an /auth-callback route

    This is the route that'll take care of exchanging the
    authorization code with a fresh access token and save it in a cookie.

    Notice that we set its name parameter: this is how we can generate
    its URL automatically with the request.url_for method.
    "

    After the user authenticates in Fief and authorizes the application,
    Fief will redirect the user to this endpoint.

    Here we get the tokens and userinfo from Fief and store the userinfo,
    then we redirect to the main page.
    """
    redirect_uri = request.url_for(
        "auth_callback"
    )  # the arg is the name of the function
    # docs:
    # We generate an access token
    #
    # We finish the OAuth2 flow by exchanging the
    # authorization code with a fresh access token.
    #
    tokens, userinfo = await fief.auth_callback(code, redirect_uri)
    # this redirects to after login
    base_url = f"{request.url.scheme}://{request.url.netloc}"
    response = RedirectResponse(base_url)
    response.set_cookie(
        SESSION_COOKIE_NAME,
        # access token includes permissions and scope
        # these can be accessed with
        # t = await fief.validate_access_token(tokens["access_token"])
        # t["permissions"]
        tokens["access_token"],
        max_age=tokens["expires_in"],
        # docs:
        #
        # Set the cookie as HTTPOnly
        # For such sensitive values, it's strongly recommended
        # to set the cookie as HTTPOnly. It means that it won't
        # be possible to read its value from JavaScript,
        # reducing potential attacks.
        httponly=True,
        # docs:
        #
        # Set the cookie as secure in production
        # For such sensitive values, it's strongly recommended
        # to set the cookie as Secure. It tells the browser to
        # send the cookie only on HTTPS (SSL) connection, reducing
        # the risk of the access token to be stolen by a attacker
        # between the client and the server.
        # However, in a local environment, you usually don't serve your
        # application with SSL. That's why we set it to False in this example.
        # A common approach to handle this is to have an environment variable to
        # control this parameter, so you can disable it in local and enable it in production.
        secure=False,
    )
    # save userinfo in memory cache
    await memory_userinfo_cache.set(uuid.UUID(userinfo["sub"]), userinfo)
    # redirect to the main page
    return response


@router.get("/logout", name="logout")
async def logout(
    request: Request, response: Response, user: UserSchema = Depends(get_current_user)):  # noqa B008
    """Logout the user
    Remove the cookie and redirect to fief logout, then to the main page
    """
    base_url = f"{request.url.scheme}://{request.url.netloc}"
    logout_url = await fief.logout_url(str(base_url))
    response = RedirectResponse(logout_url)
    response.delete_cookie(SESSION_COOKIE_NAME)
    return response


@router.get("/projects", response_model=FastUI, response_model_exclude_none=True)
async def projects(
    request: Request,
    user: Optional[UserSchema] = Depends(get_current_user),  # noqa B008
) -> list[AnyComponent]:
    webapp: TimelinkWebApp = request.app.state.webapp
    # collect the info from TimelinkWebApp
    return await projects_info(webapp, request=request, user=user)


@router.post("/projects/select", response_model=FastUI, response_model_exclude_none=True)
async def project_select(
    request: Request,
    project_name: Annotated[str, Form()],
    user: UserSchema = Depends(get_current_user),  # noqa B008
):
    webapp: TimelinkWebApp = request.app.state.webapp
    if project_name is not None:
        user.current_project_name = project_name
        users_db = webapp.users_db
        with users_db.session() as session:
            project: Project = users_db.get_project_by_name(project_name, session=session)
        user.current_project = project
        access_level: ProjectAccess = users_db.get_user_project_access(user.id, project.id, session=session)
        if access_level is None:
            raise ValueError(f"User {user.name} has no access to project {project_name}")

        project_db = TimelinkDatabase(project.databaseURL)
        user.current_project_db = project_db

    return [c.FireEvent(event=GoToEvent(url='/projects'))]


@router.get("/info", response_model=FastUI, response_model_exclude_none=True)
async def info(
    request: Request, user: UserSchema = Depends(get_current_user)  # noqa B008
) -> list[AnyComponent]:
    webapp: TimelinkWebApp = (
        request.app.state.webapp
    )  # collect the info from TimelinkWebApp
    return await webapp_info(webapp, request=request, user=user)


@router.get("/explore", response_model=FastUI, response_model_exclude_none=True)
async def explore(
    request: Request, user: UserSchema = Depends(get_current_user)  # noqa B008
) -> list[AnyComponent]:
    markdown = """\
* See list of attributes
* See list of relations
* See list of events
"""
    return await home_page(
        c.Markdown(text=markdown), request=request, title="Explore", user=user
    )


@router.get("/sources", response_model=FastUI, response_model_exclude_none=True)
async def sources(
    request: Request, user: UserSchema = Depends(get_current_user)  # noqa B008
) -> list[AnyComponent]:
    markdown = """\
* View sources
* Translate sources
* Import souces

"""
    return await home_page(
        c.Markdown(text=markdown), request=request, title="Sources", user=user
    )


@router.get("/adm", response_model=FastUI, response_model_exclude_none=True)
async def admin(
    request: Request,
    user: UserSchema = Depends(get_current_user),  # noqa B008
) -> list[AnyComponent]:
    webapp: TimelinkWebApp = request.app.state.webapp
    markdown = f"""\
### Welcome {user.name} to the admin page

Timelink administration consists of:

* Managing [authentication]({webapp.auth_manager}/admin/)
    * Who can login, with what permissions, passwords, etc.
* Managing [application resources]({webapp.app_manager})
    * Projects, associate users with projects with specific permissions.
    * General application settings, logs, etc.

"""
    return await home_page(
        c.Markdown(text=markdown), request=request, title="Admin", user=user
    )


@router.get("/", name="home", response_model=FastUI, response_model_exclude_none=True)
async def home(
    request: Request,
    user: Optional[UserSchema] = Depends(get_current_user),  # noqa B008
) -> list[AnyComponent]:
    markdown = """\

Timelink provides:

* a notation (kleio) for textual transcription of historical sources,
  capable of capturing complex information about people and their
  relations in any type of historical documents.
* a text processor that extracts information from the
  kleio transcriptions, infers personal attributes and
  relations from agent roles in historical events and generates
   data in formats suitable for database import.
* a database management system capable of processing time
   varying attributes and relations, assist in reconstructing
biographies and networks from fragmentary information.

---

    """

    return await home_page(
        c.Markdown(text=markdown),
        request=request,
        title="Welcome to Timelink",
        user=user,
    )



""" Return the attributes of a group of entities in a dataframe."""

import warnings
from sqlalchemy import column, select, not_
from sqlalchemy.orm.session import Session
import pandas as pd
from IPython.display import display
from timelink.api.database import TimelinkDatabase
from timelink.pandas.entities_with_attribute import entities_with_attribute
from timelink.pandas.styles import styler_row_colors


def group_attributes(
    group: list,
    entity_type="entity",
    include_attributes=None,
    exclude_attributes=None,
    show_elements=None,
    db: TimelinkDatabase = None,
    session: Session = None,
    sql_echo=False,
):
    """Return the attributes of a group of entities in a dataframe.

    Args:
    group: list of ids
    entity_type: type of entities to show
    show_elements: elements of entity type to include
                (e.g. name, description, obs)
    include_attributes: list of attribute types to include
    exclude_attributes: list of attribute types to exclude
    db: a TimelinkDatabase object if None specify session
    session: a sqlalchemy session object, if None specify db
    sql_echo: if True echo the sql generated

    """

    if group is None:
        group = []
        warnings.warn("No list of ids specified", stacklevel=2)
        return None

    mysession = None
    dbsystem: TimelinkDatabase = None
    if db is not None:
        dbsystem = db
        mysession = dbsystem.session()
    elif session is not None:
        mysession = session
    else:
        raise (
            Exception(
                "must call TimelinkDatabase() and pass db= "
                "or specify previously openned session with session="
            )
        )

    entity_types = db.get_models_ids()
    if entity_type not in entity_types:
        raise ValueError(f"entity_type must be one of {entity_types}")

    entity_model = db.get_model(entity_type)
    # get the columns of the entity table, check if more_info is valid
    # we need the "select" to get all the columns up in the inheritance
    id_col: column = select(entity_model).selected_columns["id"]
    entity_names = select(entity_model).selected_columns.keys()
    if show_elements is None:
        show_elements = []

    cols = [id_col]
    extra_cols = []
    # collect the column objects for the select list
    for mi in show_elements:
        if mi not in entity_names:
            raise ValueError(f"{mi} is not a valid column for {entity_type}")
        else:
            extra_cols.append(select(entity_model).selected_columns[mi])

    cols.extend(extra_cols)

    attr = db.get_table("attribute")

    cols.extend(
        [
            attr.c.the_type,
            attr.c.the_value,
            attr.c.the_date,
            attr.c.obs.label("attr_obs"),
        ]
    )

    select_entities = select(entity_model).where(id_col.in_(group))

    stmt = select_entities.join(
        attr, attr.c.entity == entity_model.id, isouter=True
    ).with_only_columns(*cols)

    # these should allow sql wild cards
    # but it is not easy in sql
    if include_attributes is not None and len(include_attributes) != 0:
        stmt = stmt.where(attr.c.the_type.in_(include_attributes))
    if exclude_attributes is not None and len(exclude_attributes) != 0:
        stmt = stmt.where(not_(attr.c.the_type.in_(exclude_attributes)))

    if sql_echo:
        print(stmt)

    with mysession as session:
        records = session.execute(stmt)
        col_names = stmt.selected_columns.keys()
        df = pd.DataFrame.from_records(records, index="id", columns=col_names)
    return df


def display_group_attributes(
    ids,
    entity_type="entity",
    header_elements=None,
    header_attributes=None,
    sort_header=None,
    sort_attributes=None,
    # These go to de_row_colors
    category="id",
    cmap_name="Pastel2",
    # these go to group_attributes
    include_attributes=None,
    exclude_attributes=None,
    db: TimelinkDatabase = None,
):
    """Display attributes of a group with header and colored rows.

    Same as group attributes but a header is displayed for each entity
    and each entity is colored. The attribute list is also colored,
    to make is clear which attributes are from which entity."""

    if header_elements is None:
        header_elements = []
    if header_attributes is None:
        header_attributes = ["%"]
        column_name = "the_type"
    else:
        column_name = header_attributes[0]

    table_cols = ["the_type", "the_value", "the_date", "attr_obs"]

    header_df = entities_with_attribute(
        header_attributes[0],
        entity_type=entity_type,
        column_name=column_name,
        show_elements=header_elements,
        more_attributes=header_attributes[1:],
        filter_by=ids,
        db=db,
    )
    # remove  entries with duplicate index from header_df

    header_df = header_df.reset_index().drop_duplicates(subset='id', keep='first').set_index('id')

    if sort_header is not None:
        header_df.sort_values(sort_header, inplace=True)

    header_df["id"] = header_df.index
    header_df.reset_index(drop=True, inplace=True)
    header_cols = header_df.columns.tolist()
    if category not in header_cols:
        header_cols = [category] + header_cols
    header_df = styler_row_colors(
        header_df[header_cols], category=category, cmap_name=cmap_name
    )
    display(header_df)

    df = group_attributes(
        ids,
        entity_type=entity_type,
        show_elements=header_elements,
        include_attributes=include_attributes,
        exclude_attributes=exclude_attributes,
        db=db,
    )
    if sort_attributes is not None:
        df.sort_values(sort_attributes, inplace=True)
    df["id"] = df.index
    df.reset_index(drop=True, inplace=True)
    table_cols = df.columns.tolist()
    if category not in table_cols:
        table_cols = [category] + table_cols
    df = styler_row_colors(df[table_cols], category="id", cmap_name=cmap_name)
    display(df)



""" Generation of networks

TODO: pass DB object instead of Engine
"""

from itertools import combinations
import networkx as nx
from sqlalchemy import text

from timelink.api.database import TimelinkDatabase


def network_from_attribute(
    attribute: str,
    mode="cliques",
    user="*none*",
    db: TimelinkDatabase = None,
    session=None,
):
    """
    Generate a network from common attribute values.

    This function will generate a network connecting the entities
    that have the same value for the attribute given in the parameter.

    Args:
        attribute (str): The name (type) of the attribute used for generating the graph.
        mode (str, optional): The topology of the generated network (see bellow).
                              Valid values are "cliques" and "value-node". Defaults to "cliques".
        user (str, optional): Use real persons identified by this user. Defaults to "*none*".
        db (TimelinkDatase, optional): The TimelinkDatase object representing the target database. Defaults to None.
        session (object, optional): The session object for the database connection. Defaults to None.

    Raises:
        None

    Topology the generated network:
        * If mode = "cliques" all the entities with attribute
          will be nodes in the graph and edges will be created between
          the entities with the same value of the attribute.
        * If mode = "value-node" a node will be created for each
          different value of the attribute a and edges will be
          created linking that node to the entities which have
          that value in attribute a.
        * In both cases entities with several values for the attribute
          contribute to the overall connectivity of the graph, by
          linking clusters of same value entities.

    Returns:
        networkx.classes.graph.Graph: The generated network as a networkx Graph object.

    Each node will have an associated dictionary of attributes:

        * "id": (entity id)
        * "type": "value_node" or entity class in the database.
        * "desc": a description of the node, automatically fetched
                  from the database (names of persons for instance)
        * "is_real": a flag stating if the "id" key refers to a real
                  entity or to an occurrence.
        * "url": a link to the entity information in the database
                 in the format http://localhost:8080/mhk/database/id/entityID

    Each edge will have associated the following key-value pairs:

        * "date1": date of the atribute in the left most node
        * "date2": date of the attribute in the right most node
        * "attribute": the type of the attribute
        * "value": the value of the attribute

    Examples:

        ``G = network_from_attribute("graduated_at", mode="value-node")``

    """

    sql = (
        "select distinct the_value from attributes "
        "where the_type = :the_type and the_value <> '?'"
    )
    G = nx.Graph()
    if db is not None:
        mysession = db.session()
    elif session is not None:
        mysession = session
    else:
        raise ValueError(
            "No database nor session. Specifcy db=TimeLinkDatabase() or "
            "session=database session."
        )
    with mysession:
        result = mysession.execute(text(sql), [{"the_type": attribute}])
        values = result.all()
        for (avalue,) in values:
            if user == "*none*":
                sql = "select id,name,the_date from nattributes "
                "where the_type=:the_type and the_value = :the_value"
            else:
                sql = "SELECT IFNULL( (select rid from rlinks where instance=n.id"
                " and user=:user),id) as id, name, the_date  from nattributes n "
                "where the_type=:the_type and the_value = :the_value"
            # TODO: also fetch the instance SELECT IFNULL( (select rid from rlinks
            #       where instance=n.id and user=:user),id) as id, id as instance, name,..
            #       then test if id=instance. If not add attribute to node is_real=yes
            #       otherwise "no". do the same with the first select.
            # TODO add to nodes a url attribute if host and dbase are present
            #        is present https://joaquims-mbpr.local/mhk/toliveira/id/rp-1
            result = mysession.execute(
                text(sql), [{"the_type": attribute, "the_value": avalue, "user": user}]
            )
            entities = result.all()

            if mode == "value-node":
                G.add_node(avalue, desc=avalue, type=attribute)
                for id, name, date in entities:
                    G.add_node(id, desc=name, type="person")
                    G.add_edge(
                        avalue,
                        id,
                        date1=date,
                        date2=date,
                        attribute=attribute,
                        value=avalue,
                    )
            elif len(entities) > 1:
                for id, name, _date in entities:
                    G.add_node(
                        id, desc=name, type="person"
                    )  # this should come from the entity class
                pairs = list(combinations(entities, 2))
                # TODO: optional date range filtering
                for (e1, _n1, d1), (e2, _n1, d2) in pairs:
                    G.add_edges_from(
                        [
                            (
                                e1,
                                e2,
                                {
                                    "date1": d1,
                                    "date2": d2,
                                    "attribute": attribute,
                                    "value": avalue,
                                },
                            )
                        ]
                    )
    return G



""" Test timelink Pandas integrations"""

import pytest
from pathlib import Path

from timelink.api.database import (
    TimelinkDatabase,
    is_postgres_running,
    get_postgres_container_user,
    get_postgres_container_pwd,
)
from timelink.kleio.importer import import_from_xml
from timelink.api.models import base  # pylint: disable=unused-import. # noqa: F401
from tests import TEST_DIR

from timelink.pandas import pname_to_df, entities_with_attribute, attribute_values

# pytestmark = skip_on_travis


@pytest.fixture(scope="module")
def dbsystem(request):
    """Create a database for testing"""
    db_user = None
    db_pwd = None
    db_type, db_name, db_url, db_user, db_pwd = request.param
    if db_type == "postgres" and is_postgres_running():
        db_user = get_postgres_container_user()
        db_pwd = get_postgres_container_pwd()
    database = TimelinkDatabase(db_name, db_type, db_url, db_user, db_pwd)
    db = database.session()
    """Load sample data in the test database"""
    file1: Path = Path(TEST_DIR, "xml_data/b1685.xml")
    file2: Path = Path(TEST_DIR, "xml_data/dehergne-a.xml")
    try:
        import_from_xml(file1, session=db)
        import_from_xml(file2, session=db)
    except Exception as exc:
        print(exc)
        raise
    try:
        yield database
    finally:
        database.drop_db(db)
        db.close()


@pytest.mark.parametrize(
    "dbsystem",
    [
        # db_type, db_name, db_url, db_user, db_pwd
        ("sqlite", ":memory:", None, None, None),
        ("postgres", "tests", None, None, None),
    ],
    indirect=True,
)
def test_name_to_df(dbsystem):
    """test generation of dataframe with people with a certain name"""

    # test finding by similarity
    names_df = pname_to_df("matias carvalho", db=dbsystem, similar=True, sql_echo=True)
    print(names_df)
    assert names_df is not None, "name_to_df returned None"
    assert len(names_df) > 0

    # test finding by exact match
    names_df = pname_to_df("matias carvalho", sql_echo=True, db=dbsystem, similar=False)
    assert names_df is None, "name_to_df returned results but should not have"
    print(names_df)


@pytest.mark.parametrize(
    "dbsystem",
    [
        # db_type, db_name, db_url, db_user, db_pwd
        ("sqlite", ":memory:", None, None, None),
        ("postgres", "tests", None, None, None),
    ],
    indirect=True,
)
def test_attribute_values(dbsystem):
    """test generation of dataframe with people/objects with a certain attribute"""

    locais_residencia = attribute_values("residencia", db=dbsystem, sql_echo=True)
    assert locais_residencia is not None, "attribute_values returned None"


@pytest.mark.parametrize(
    "dbsystem",
    [
        # db_type, db_name, db_url, db_user, db_pwd
        ("sqlite", ":memory:", None, None, None),
        ("postgres", "tests", None, None, None),
    ],
    indirect=True,
)
def test_entities_with_attribute(dbsystem):
    """Test generation of dataframe from attributes"""
    df = entities_with_attribute(
        db=dbsystem,
        entity_type="person",
        the_type="residencia",
        the_value="soure",
        show_elements=["name", "sex"],
        more_attributes=[],
        sql_echo=True,
    )
    assert df is not None, "entities_with_attribute returned None"
    assert len(df) > 0, "entities_with_attribute returned empty dataframe"
    print(df)


@pytest.mark.parametrize(
    "dbsystem",
    [
        # db_type, db_name, db_url, db_user, db_pwd
        ("sqlite", ":memory:", None, None, None),
        ("postgres", "tests", None, None, None),
    ],
    indirect=True,
)
def test_entities_with_attribute_list(dbsystem):
    """Test generation of dataframe from attributes"""
    df = entities_with_attribute(
        db=dbsystem,
        entity_type="person",
        the_type=["nascimento", "entrada", "partida", "estadia", "morte"],
        show_elements=["name", "sex"],
        more_attributes=[],
        sql_echo=True,
    )
    assert df is not None, "entities_with_attribute_list returned None"
    assert len(df) > 0, "entities_with_attribute_lisy returned empty dataframe"
    print(df)



from typing import List, Optional
from datetime import datetime
import json

# for sqlalchemy 2.0 ORM
# see https://docs.sqlalchemy.org/en/20/orm/declarative_config.html
# and https://docs.sqlalchemy.org/en/20/orm/declarative_tables.html#orm-declarative-table

from sqlalchemy import ForeignKey  # pylint: disable=import-error
from sqlalchemy import String  # pylint: disable=import-error
from sqlalchemy import Integer  # pylint: disable=import-error
from sqlalchemy import DateTime  # pylint: disable=import-error
from sqlalchemy.orm import backref  # pylint: disable=import-error
from sqlalchemy.orm import Mapped  # pylint: disable=import-error
from sqlalchemy.orm import mapped_column  # pylint: disable=import-error
from sqlalchemy.orm import relationship  # pylint: disable=import-error

from timelink.kleio.utilities import kleio_escape
from .base_class import Base


class Entity(Base):
    """ORM Model root of the object hierarchy.

     All entities in a Timelink/MHK database have an entry in this table.
     Each entity is associated with a class that allow access to a
     specialization table with more columns for that class.

    This corresponds to the model described as "Joined Table Inheritance"
    in sqlalchemy (see https://docs.sqlalchemy.org/en/14/orm/inheritance.html)

    TODO: specialize in TemporalEntity to implement https://github.com/time-link/timelink-kleio/issues/1
         Acts, Sources, Attributes and Relations are TemporalEntities
    """

    __tablename__ = "entities"
    __allow_unmapped__ = True

    #: str: unique identifier for the entity
    id: Mapped[str] = mapped_column(primary_key=True)
    #: str: name of the class. Links to pom_som_mapper class
    # TODO: https://docs.sqlalchemy.org/en/20/orm/declarative_config.html
    pom_class: Mapped[str] = mapped_column(
        "class",
        String,
        # ForeignKey('classes.id', name='fk_entity_class', use_alter=True),
        index=True,
    )
    #: str: id of the entity inside which this occurred.
    inside: Mapped[Optional[str]] = mapped_column(
        String, ForeignKey("entities.id", ondelete="CASCADE"), index=True
    )
    #: int: sequential order of this entity in the source
    the_order = mapped_column(Integer, nullable=True)
    #: int: the nesting level of this entity in the source
    the_level = mapped_column(Integer, nullable=True)
    #: int: line in which the entity occurred in the source
    the_line = mapped_column(Integer, nullable=True)
    #: str: name of the kleio group that produced this entity
    groupname = mapped_column(String, index=True, nullable=True)
    #: datetime: when this entity was updated in the database
    updated = mapped_column(DateTime, default=datetime.utcnow, index=True)
    #: datetime: when this entity was added to the full text index
    indexed = mapped_column(DateTime, index=True, nullable=True)

    # These are defined in relation.py
    # rels_in = relationship("Relation", back_populates="dest")
    # rels_out = relationship("Relation", back_populates="org")

    # this based on
    # https://stackoverflow.com/questions/28843254
    #: list(Entity): list of Entity objects contained in this entity
    contains: Mapped[List["Entity"]] = relationship(
        "Entity",
        backref=backref("contained_by", remote_side="Entity.id"),
        cascade="all",
    )

    # see https://docs.sqlalchemy.org/en/14/orm/inheritance.html
    # To handle non mapped pom_class
    #      see https://github.com/sqlalchemy/sqlalchemy/issues/5445
    #
    #    __mapper_args__ = {
    #       "polymorphic_identity": "entity",
    #    "polymorphic_on": case(
    #        [(type.in_(["parent", "child"]), type)], else_="entity"
    #    ),
    #
    #  This defines what mappings do exist
    # [aclass.__mapper_args__['polymorphic_identity']
    #              for aclass in Entity.__subclasses__()]

    __mapper_args__ = {
        "polymorphic_identity": "entity",
        "polymorphic_on": pom_class,
    }

    @classmethod
    def get_subclasses(cls):
        """Get the subclasses of Entity"""
        for subclass in cls.__subclasses__():
            yield from subclass.get_subclasses()
            yield subclass

    @classmethod
    def get_orm_entities_classes(cls):
        """Currently defined ORM classes that extend Entity
         (including Entity itself)


        Returns:
             list: List of ORM classes
        """
        sc = list(Entity.get_subclasses())
        sc.append(Entity)
        return sc

    @classmethod
    def get_som_mapper_ids(cls):
        """Ids of SomPomMapper references by orm classes

        Returns:
             List[str]: List of strings
        """
        return [
            aclass.__mapper_args__["polymorphic_identity"]
            for aclass in Entity.get_orm_entities_classes()
        ]

    @classmethod
    def get_tables_to_orm_as_dict(cls):
        """
        Return a dict with table name as key and ORM class as value
        """
        return {
            ormclass.__mapper__.local_table.name: ormclass
            for ormclass in Entity.get_orm_entities_classes()
        }

    @classmethod
    def get_som_mapper_to_orm_as_dict(cls):
        """
        Return a dict with pom_class id as key and ORM class as value
        """
        sc = Entity.get_orm_entities_classes()
        return {ormclass.__mapper__.polymorphic_identity: ormclass for ormclass in sc}

    @classmethod
    def get_orm_for_table(cls, table: String):
        """
        Entity.get_orm_for_table("acts")

        will return the ORM class handling the "acts" table
        """
        return cls.get_tables_to_orm_as_dict().get(table, None)

    @classmethod
    def get_orm_for_pom_class(cls, pom_class: str):
        """
        Entity.get_orm_for_pom_class("act")

        will return the ORM class corresponding to the pom_class "act"
        """
        return cls.get_som_mapper_to_orm_as_dict().get(pom_class, None)

    @classmethod
    def get_entity(cls, id: str, session=None):
        """
        Get an Entity from the database. The object returned
        will be of the ORM class defined by mappings.
        :param id: id of the entity
        :param session: current session
        :return: an Entity object of the proper class for the mapping
        """
        entity = session.get(Entity, id)
        if entity is not None:
            if entity.pom_class != "entity":
                orm_class = Entity.get_orm_for_pom_class(entity.pom_class)
                object_for_id = session.get(orm_class, id)
                return object_for_id
            else:
                return entity
        else:
            return None

    def get_extra_info(self):
        """Return a dictionatry with extra information about this entity or None

        if entity has an 'obs' field and that field
        contains the string 'extra_info:' then the rest
        of the field is considered a json representation of
        extra information about the entity. Extra information can
        be comments and original wording of field values.

        This method returns
        the json information as a dictionnary.

        Currently the dictionary has the following structure:

            {'field_name': {'comment': text_of_comment, 'original': original_wording}}
        """
        obs = getattr(self, 'obs', None)
        if obs is not None:
            s = obs.split('extra_info:')
            if len(s) > 1:
                exs = s[1].strip()
                extra_info = json.loads(exs)
            else:
                extra_info = None
        else:
            extra_info = None
        return extra_info

    def __repr__(self):
        return (
            f'Entity(id="{self.id}", '
            f'pom_class="{self.pom_class}",'
            f'inside="{self.inside}", '
            f"the_order={self.the_order}, "
            f"the_level={self.the_level}, "
            f"the_line={self.the_line}, "
            f'groupname="{self.groupname}", '
            f"updated={self.updated}, "
            f"indexed={self.indexed},"
            f")"
        )

    def __str__(self):
        return f"{self.groupname}${kleio_escape(self.id)}/type={kleio_escape(self.pom_class)}"

    def to_kleio(self, ident="", ident_inc="  "):
        s = f"{ident}{str(self)}"
        contained_entities: List[Entity] = self.contains
        # sort by the_order
        contained_entities.sort(key=lambda x: x.the_order)
        for inner in contained_entities:
            innerk = inner.to_kleio(ident=ident + ident_inc, ident_inc=ident_inc)
            if innerk != "":
                s = f"{s}\n{innerk}"
        return s



from typing import Optional
from fastapi import Request
from fastui import components as c
from pydantic import BaseModel, Field

from timelink.app.backend.timelink_webapp import TimelinkWebApp
from timelink.app.schemas.user import UserSchema, UserPropertySchema
from .home_page import home_page


# class to hold the Webapp info
class WebAppInfo(BaseModel):
    """Classe to holf the WebApp info."""

    info_label: str = Field(title="Label")
    info_value: Optional[str] = Field(title="Value", default="<NA>")


async def webapp_info(
    webapp: TimelinkWebApp, request: Request, user: UserSchema
) -> c.Page:
    info_list = [
        WebAppInfo(info_label=label, info_value=value)
        for (label, value) in webapp.get_info().items()
    ]
    if user is not None:
        user_properties = user.properties
    else:
        user_properties = []

    return await home_page(
        c.Page(components=[
            c.Heading(text="User Info", level=4),
            c.Table(data=user_properties, data_model=UserPropertySchema),
            c.Heading(text="WebApp Info", level=4),
            c.Table(data=info_list)]),
        request=request,
        title="Info",
        user=user,
    )



# pylint: disable=import-error

from sqlalchemy import Column, String, ForeignKey, Index
from sqlalchemy.orm import relationship

from timelink.kleio.utilities import kleio_escape, quote_long_text
from timelink.mhk.models.entity import Entity


class Relation(Entity):
    """Represents a relation between two entities."""

    __allow_unmapped__ = True
    __tablename__ = "relations"
    __table_args__ = {"extend_existing": True}

    id = Column(String, ForeignKey("entities.id"), primary_key=True)
    # rel_entity = relationship("Entity",
    #                   foreign_keys='id',back_populates='rel')
    origin = Column(String, ForeignKey("entities.id"), index=True)
    # org = relationship(Entity, foreign_keys=[origin],
    #                   back_populates='rels_out')

    destination = Column(String, ForeignKey("entities.id"), index=True)
    # dest = relationship("Entity", foreign_keys=[destination],
    #                    back_populates="rels_in")
    the_type = Column(String, index=True)
    the_value = Column(String, index=True)
    the_date = Column(String, index=True)
    obs = Column(String)

    __mapper_args__ = {
        "polymorphic_identity": "relation",
        "inherit_condition": id == Entity.id,
    }

    __table_args__ = (Index("relations_type_value", "the_type", "the_value"),)

    def __repr__(self):
        sr = super().__repr__()
        return (
            f"Relation(id={sr}, "
            f'origin="{self.origin}", '
            f'destination="{self.destination}", '
            f'the_type="{self.the_type}", '
            f'the_value="{self.the_value}", '
            f'the_date="{self.the_date}"", '
            f"obs={self.obs}"
            f")"
        )

    def __str__(self):
        if (
            self.dest is not None and self.dest.pom_class == "person"
        ):  # pylint: disable=no-member
            r = (
                f"rel${kleio_escape(self.the_type)}/"
                f"{quote_long_text(self.the_value)}/"
                f"{kleio_escape(self.dest.name)}"  # pylint: disable=no-member
                f"/{self.destination}/{self.the_date}"
            )
        else:
            r = (
                f"rel${self.the_type}/{quote_long_text(self.the_value)}/"
                f"{self.destination}/{self.the_date}"
            )
        if self.obs is not None:
            r = f"{r}/obs={quote_long_text(self.obs)}"
        return r


Entity.rels_out = relationship(
    "Relation", foreign_keys=[Relation.origin], backref="dest"
)

Entity.rels_in = relationship(
    "Relation", foreign_keys=[Relation.destination], backref="org"
)



"""
Pandas related utilities for Timelink
In progress. Originally developped for the FAUC1537-1919 project.
Now being integrated in the timelink-py package
"""
# Utilities for interacting with pandas
#
# IDEAS:
#    baptisms=function_to_df( function='n',
#                               act_tpe='bap',
#                               column_name='child', # instead of function
#                               person_info=True, # adds function_name, function_sex,
#                                                 #  columns
#                               dates_in=(after,before),
#                               name_like = name of person with function
#                               more_funcs=['pn','mn','ppn','mpn','pmn','mmn'],....)

from .name_to_df import pname_to_df  # noqa: F401
from .entities_with_attribute import entities_with_attribute  # noqa: F401
from .attribute_values import attribute_values  # noqa: F401
from .group_attributes import group_attributes  # noqa: F401
from .group_attributes import display_group_attributes  # noqa: F401
from .styles import category_palette, styler_row_colors  # noqa: F401



""" Main class for the Timelink web application. """

import os
import json
from typing import List

import pandas
from sqlalchemy import select
from sqlalchemy.orm import selectinload
from sqlalchemy.engine.url import make_url

import timelink
from timelink.api.database import get_postgres_dbnames, get_sqlite_databases
from timelink.app.schemas.project import ProjectSchema
from timelink.kleio.kleio_server import KleioServer
from timelink.app.models import UserDatabase, User, UserProperty  # noqa
from timelink.app.models.project import Project


class TimelinkWebApp:
    """A class to interact with the Timelink system
    from a FastAPI web application

    It stores TimelinkDatabase, KleioServer objects
    and Fief (user management) objects.

    Attributes:
        app_name (str): Name of the application.
        timelink_home (str): Directory where the Timelink database is located.
        host_url (str): URL of the Timelink web application.
        kleio_server (KleioServer): A KleioServer instance.
        users_db_type (str): Type of the users database (sqlite or postgres).
        users_db_name (str): Name of the users database.
        users_db (UserDatabase): A UserDatabase instance.
        auth_manager (str): URL of the authentication manager.
        app_manager (str): URL of the application manager.
        kleio_image (str): Name of the Kleio image to use.
        postgres_image (str): Name of the postgres image to use.
        postgres_version (str): Version of the postgres image to use.
        sqlite_dir (str): Directory where the sqlite databases are located.
        stop_duplicates (bool): If True, stop other kleio servers for the same timelink home.

    """
    # this should be set in a Dependency
    after_auth_url = None
    # Url in fief to authenticate
    # must be set with fief.auth_url(redirect_uri="http://localhost:8000")
    # see https://fief-dev.github.io/fief-python/fief_client.html#Fief.auth_url
    auth_url = None
    after_logout_url = None
    # Url in fief to logout
    # must be set with fief.logout_url(redirect_uri="http://localhost:8000")
    # see https://fief-dev.github.io/fief-python/fief_client.html#Fief.logout_url
    logout_url = None

    def __init__(
        self,
        app_name: str = "timelink",
        timelink_url: str = "http://localhost:8008",
        auth_manager: str = "http://localhost:8000",
        app_manager: str = "http://localhost:8008/admin/",
        timelink_home: str = None,
        kleio_server: KleioServer = None,
        users_db_type: str = "sqlite",
        users_db_name: str = "timelink_users.sqlite",
        kleio_image=None,
        kleio_version=None,
        kleio_token=None,
        kleio_update=False,
        postgres_image=None,
        postgres_version=None,
        sqlite_dir=None,
        stop_duplicates=True,  # kleio server duplicates
        initial_users: list[User] = None,
        **connection_args,
    ):
        """Create a TimelinkWebApp instance

        Setup of Kleio Server and Timelink
        database is done here.

        Several functions are provided to
        manage the kleio files and access the database.

        Args:
            app_name: name of the application
            timelink_url: URL of the Timelink web application
            timelink_home: directory where the Timelink database is located
            kleio_server: a KleioServer instance
            users_db_type: type of the users database (sqlite or postgres)
            users_db_name: name of the users database
            kleio_image: name of the Kleio image to use
            kleio_version: version of the Kleio image to use
            kleio_token: token to access the Kleio server
            kleio_update: if True, update the Kleio server
            postgres_image: name of the postgres image to use
            postgres_version: version of the postgres image to use
            sqlite_dir: directory where the sqlite databases are located
            initial_users: list of initial users (deprecated)
            stop_duplicates: if True, stop duplicates
            **connection_args: extra arguments to pass to the TimelinkDatabase

        Returns:
            A TimelinkWebApp instance
        """
        self.app_name = app_name
        self.timelink_home = timelink_home
        self.host_url = timelink_url
        self.kleio_server = kleio_server
        self.kleio_version = kleio_version
        self.users_db_type = users_db_type
        self.users_db_name = users_db_name
        self.users_db = None
        self.auth_manager = auth_manager
        self.app_manager = app_manager
        self.kleio_image = kleio_image
        self.postgres_image = postgres_image
        self.postgres_version = postgres_version
        self.sqlite_dir = sqlite_dir
        self.stop_duplicates = stop_duplicates
        # deprecated
        self.initial_users = initial_users
        #
        self.kleio_token = kleio_token
        self.kleio_update = kleio_update
        self.projects: List[ProjectSchema] = []

        if initial_users is None:
            self.initial_users = []
        if self.timelink_home is None:
            self.timelink_home = KleioServer.find_local_kleio_home()
        if self.users_db_type == "sqlite":
            if self.sqlite_dir is None:
                self.sqlite_dir = os.path.join(self.timelink_home, "system/db/sqlite")
            if not os.path.exists(self.sqlite_dir):
                os.makedirs(self.sqlite_dir)
            self.users_db = UserDatabase(
                db_type=self.users_db_type,
                db_name=self.users_db_name,
                db_path=self.sqlite_dir,
                stop_duplicates=self.stop_duplicates,
                initial_users=self.initial_users,
                **connection_args,
            )
        elif self.users_db_type == "postgres":
            self.users_db = UserDatabase(
                db_type=self.users_db_type,
                db_name=self.users_db_name,
                postgres_image=self.postgres_image,
                postgres_version=self.postgres_version,
                stop_duplicates=self.stop_duplicates,
                initial_users=self.initial_users,
                **connection_args,
            )
        else:
            raise ValueError(f"Invalid database type: {self.users_db_type}")

        if self.kleio_server is not None:
            self.kleio_server = kleio_server
        else:
            if self.timelink_home is not None:
                self.kleio_server: KleioServer = KleioServer.start(
                    kleio_home=self.timelink_home,
                    kleio_image=self.kleio_image,
                    kleio_version=self.kleio_version,
                    kleio_admin_token=self.kleio_token,
                    update=self.kleio_update,
                    stop_duplicates=self.stop_duplicates,
                )
        self.update_projects()

    def get_info(self, show_token=False, show_password=False):
        """Print information about the Timel8nk Webapp object"""
        if not show_password:
            # mask any password that might be present in the dabase URL
            url = make_url(str(self.users_db.engine.url))
            if url.password:
                url.password = '****'
            db_url = str(url)
        else:
            db_url = str(self.users_db.engine.url)

        info_dict = {
            "Timelink version": timelink.version,
            "Timelink home": self.timelink_home,
            "Timelink host URL": self.host_url,
            "Timelink users database": db_url,
            "Kleio server": self.kleio_server.get_url(),
            "Kleio version requested": self.kleio_version,
            "SQLite directory": self.sqlite_dir,
            "Postgres image": self.postgres_image,
            "Postgres version": self.postgres_version,
        }

        kserver: KleioServer = self.kleio_server
        if kserver is not None:
            info_dict.update(
                {
                    "Kleio server token": kserver.get_token(),
                    "Kleio server URL": kserver.get_url(),
                    "Kleio server home": kserver.get_kleio_home(),
                }

            )
            if not show_token:
                info_dict["Kleio server token"] = kserver.get_token()[:5] + "..."
            if kserver.container is not None:
                info_dict["Kleio server container"] = kserver.container.name
            info_dict["Kleio version requested"] = self.kleio_version
            labels = kserver.container.labels
            build = labels.get("BUILD", "")
            version = labels.get("VERSION", "")
            build_date = labels.get("BUILD_DATE", "")
            if version != "":
                info_dict["Kleio server version"] = f"{version}.{build} ({build_date})"
        if self.users_db_type == "sqlite":
            info_dict["SQLite directory"] = self.sqlite_dir
        elif self.users_db_type == "postgres":
            info_dict.update(
                {
                    "Postgres image": self.postgres_image,
                    "Postgres version": self.postgres_version,
                    "Postgres user": self.db.db_user,
                    "Postgres password": self.db.db_pwd,
                }
            )
            if not show_password:
                info_dict["Postgres password"] = "..."
        return info_dict

    def get_project_dirs(self):
        """Get the list of projects

        Projects are sub directories of the
        timelink home directory / projects directory."""
        projects = []
        # get the sub directories of timelink-home/projects
        projects_dir = os.path.join(self.timelink_home, "projects")
        if os.path.exists(projects_dir):
            projects = [
                d
                for d in os.listdir(projects_dir)
                if os.path.isdir(os.path.join(projects_dir, d))
            ]
        return projects

    def update_projects(self) -> List[Project]:
        """Get the list of projects

        Get the list of projects from the subdirectories
        of the "projects" directory in the Timelink home directory.

        Check the database for projects entries and merge the two lists
        so that the database has the most recent information.
        """
        if self.timelink_home is None:
            return []
        with self.users_db.session() as session:
            projs = session.scalars(select(Project).options(selectinload(Project.users))).all()
            if projs is not None:
                self.projects = [ProjectSchema.model_validate(proj) for proj in projs]
            else:
                self.projects = []
            existing_project_names = [p.name.upper() for p in self.projects]
            pdirs = self.get_project_dirs()
            for pdir in pdirs:
                if pdir.upper() not in existing_project_names:
                    # todo: check if there is a project settings in the dir
                    project = Project(name=pdir)
                    session.add(project)
                    self.projects.append(project)
            session.commit()
        return self.projects

    def print_info(self):
        info_dict = self.get_info()
        print(json.dumps(info_dict, indent=4))
        print(self.__repr__())

    def get_imported_files(self, data_frame=True, **kwargs):
        """Get the list of imported files in the database

        See the get_imported_files method in the TimelinkDatabase class:
        :meth:`timelink.api.database.TimelinkDatabase.get_imported_files`

        Args:
            data_frame: if True, return a pandas DataFrame; otherwise,
                        return a list of dictionaries
            **kwargs: extra arguments to pass to the get_imported_files method
        """
        ifiles = self.db.get_imported_files(**kwargs)

        if data_frame:
            if len(ifiles) == 0:
                return pandas.DataFrame()
            ifiles_json = [f.model_dump() for f in ifiles]
            ifiles_df = pandas.DataFrame(ifiles_json)
            ifiles_df["nerrors"] = ifiles_df["nerrors"].astype("Int64")
            ifiles_df["nwarnings"] = ifiles_df["nerrors"].astype("Int64")
            return ifiles_df
        else:
            return ifiles

    def update_from_sources(self, **kwargs):
        """Update the database from a list of sources

        See the update_from_sources method in the TimelinkDatabase class:
        :meth:`timelink.api.database.TimelinkDatabase.update_from_sources`

        """
        self.db.update_from_sources(**kwargs)

    def get_import_status(self, data_frame=True, **kwargs):
        """Get the import status of Kleio Files

        Returns:
            A dictionary with the status of the import process
        """
        ifiles = [f.model_dump() for f in self.db.get_import_status(**kwargs)]
        if data_frame:
            if len(ifiles) == 0:
                return pandas.DataFrame()
            # create a pandas Data frame
            ifiles_df = pandas.DataFrame(ifiles)
            # convert the column "status" to the enum value
            ifiles_df["status"] = ifiles_df["status"].apply(lambda x: x.value)
            ifiles_df["import_status"] = ifiles_df["import_status"].apply(
                lambda x: x.value
            )
            # convert the column "import_errors" to int with NA as 0
            # https://stackoverflow.com/questions/21287624/convert-pandas-column-containing-nans-to-dtype-int
            ifiles_df["import_errors"] = ifiles_df["import_errors"].astype("Int64")
            ifiles_df["import_warnings"] = ifiles_df["import_errors"].astype("Int64")
            return ifiles_df.fillna(0)
        else:
            return ifiles

    def get_sqlite_databases(self, sqlite_dir=None, **kwargs):
        """Get the list of sqlite databases

        Args:
            sqlite_dir: directory where the sqlite databases are located
            **kwargs: extra arguments to pass to the get_sqlite_databases function

        Returns:
            A list of sqlite databases
        """
        if sqlite_dir is None:
            sqlite_dir = self.sqlite_dir
        return get_sqlite_databases(directory_path=sqlite_dir, **kwargs)

    def get_postgres_databases(self):
        """Get the list of postgres databases

        Returns:
            A list of postgres databases
        """
        return get_postgres_dbnames()

    def table_row_count_df(self):
        """Return the row count of all tables in the database"""
        tables = self.db.table_row_count()
        tables_df = pandas.DataFrame(tables, columns=["table", "count"])
        return tables_df

    def get_file_paths(self, file_spec, rows, column):
        """Get the file paths from DataFrame of from a string

        TODO: #27 add parameter to convert the paths to absolute local paths"""
        if isinstance(file_spec, pandas.DataFrame):
            if column not in file_spec.columns:
                raise Exception(f"There is no {column} in the DataFrame")
            if rows is None:
                raise Exception("The 'rows' argument must be present")
            if type(rows) is not list:
                rows = [rows]
            file_paths = file_spec.iloc[list(rows)][column].tolist()
            return file_paths
        else:
            return []

    def get_import_rpt(
        self, file_spec: pandas.DataFrame | str, rows=None, match_path=False, **kwargs
    ):
        """Show the import report for a given file specification

        Args:
            file_spec: file specification (DataFrame or string)
                       If a DataFrame, it should have the columns 'path'
                       and the arguments 'rows' must be present
            rows: if file_spec is a DataFrane, the row number to show
            match_path: if True, the path is used to retrieve the import report;
                        if false the filename is used (default).
            **kwargs: extra arguments to pass to the show_import_rpt method
                      in the TimelinkDatabase class

        """
        rpt = ""
        if match_path:
            column = "path"
        else:
            column = "name"
        if isinstance(file_spec, pandas.DataFrame):
            paths = self.get_file_paths(file_spec, rows, column)
            for file in paths:
                rpt += self.db.get_import_rpt(file, match_path=match_path, **kwargs)
        elif isinstance(file_spec, str):
            return self.db.get_import_rpt(file_spec, match_path=match_path, **kwargs)
        else:
            raise ValueError
        return rpt

    def get_translation_report(self, file_spec, rows=None):
        """Show the translation report for a given file specification

        Args:
            file_spec: file specification (DataFrame or string)
                       If a DataFrame, it should have the columns 'rpt_url'
                       and the arguments 'rows' must be present
            rows: if file_spec is a DataFrane, the row number of interest
        """
        rpt = ""
        if isinstance(file_spec, pandas.DataFrame):
            if rows is None:
                raise ValueError(
                    "The 'rows' argument must be present "
                    "if the file_spec is a DataFrame"
                )
            elif type(rows) is not list:
                rows = [rows]
            if len(rows) == 0:
                raise ValueError(
                    "The 'rows' argument must be a non-empty list, or an integer"
                )

            paths = self.get_file_paths(file_spec, rows, "rpt_url")
            for file in paths:
                rpt += self.kleio_server.get_report(file)
        elif isinstance(file_spec, str):
            return self.kleio_server.get_report(file_spec)
        else:
            raise ValueError
        return rpt

    def get_kleio_files(self, data_frame=True, **kwargs):
        """Get the list of files in the kleio server.

        Alias to :meth:`timelink.notebooks.TimelinkNotebook.get_import_status`
        but returns a subset of the columns.

            #   Column              Non-Null Count  Dtype
            ---  ------              --------------  -----
            0   path                3 non-null      object
            1   name                3 non-null      object
            2   size                3 non-null      int64
            3   directory           3 non-null      object
            4   modified            3 non-null      datetime64[ns, UTC]
            5   modified_iso        3 non-null      datetime64[ns, UTC]
            6   modified_string     3 non-null      object
            7   qtime               3 non-null      datetime64[ns, UTC]
            8   qtime_string        3 non-null      object
            9   source_url          3 non-null      object
            10  status              3 non-null      object
            11  translated          3 non-null      datetime64[ns, UTC]
            12  translated_string   3 non-null      object
            13  errors              3 non-null      int64
            14  warnings            3 non-null      int64
            15  version             3 non-null      object
            16  rpt_url             3 non-null      object
            17  xml_url             3 non-null      object
            18  import_status       3 non-null      object
            19  import_errors       3 non-null      Int64
            20  import_warnings     3 non-null      Int64
            21  import_error_rpt    3 non-null      object
            22  import_warning_rpt  3 non-null      object
            23  imported            3 non-null      int64
            24  imported_string     3 non-null      int64
        """
        result = self.get_import_status(**kwargs)
        return result[
            [
                "path",
                "name",
                "modified",
                "status",
                "translated",
                "errors",
                "warnings",
                "import_status",
                "import_errors",
                "import_warnings",
                "import_error_rpt",
                "import_warning_rpt",
                "imported",
                "rpt_url",
                "xml_url",
            ]
        ]



""" Utilities for acessing timelink from Jupyter notebooks.

.. codeauthor:: Joaquim Ramos de Carvalho <joaquimcarvalho@mpu.edu.mo>
"""

from .timelink_notebook import TimelinkNotebook   # noqa
from timelink.api.database import get_postgres_dbnames  # noqa
from timelink.api.database import get_sqlite_databases  # noqa
from timelink.api.database import TimelinkDatabase  # noqa
from timelink.api.database import is_valid_postgres_db_name  # noqa
from timelink.kleio import KleioServer  # noqa
from timelink.mhk.models import base  # noqa



""" Models for system tables

Tables defined here:

    - syspar: system parameters
    - syslog: system log

"""

from enum import Enum
from typing import Optional

from datetime import datetime
from pydantic import BaseModel, ConfigDict  # pylint: disable=import-error

from sqlalchemy import Column, String, Integer, DateTime  # pylint: disable=import-error
from sqlalchemy.sql import func  # pylint: disable=import-error
from sqlalchemy.orm import mapped_column  # pylint: disable=import-error
from sqlalchemy.orm import Mapped  # pylint: disable=import-error
from .base_class import Base


class SysPar(Base):
    """System parameters table"""

    __tablename__ = "syspar"
    pname = Column(String, primary_key=True, index=True)
    pvalue = Column(String)
    ptype = Column(String)
    obs = Column(String)


class ParTypeSchema(str, Enum):
    """Parameter type

    Fields:
        string: string
        integer: integer
        float: float
        date: date
        boolean: boolean
        list: list
    """

    string = "string"
    integer = "integer"
    float = "float"
    date = "date"
    boolean = "boolean"
    list = "list"


class SysParSchema(BaseModel):
    """System parameters in the Timelink app

    Fields:
        pname: parameter name
        pvalue: parameter value
        ptype: parameter type
        obs: parameter description
    """

    pname: str
    pvalue: str
    ptype: ParTypeSchema
    obs: str

    model_config = ConfigDict(from_attributes=True)


class LogLevel(str, Enum):
    """Log level

    Fields:
        debug: debug
        info: info
        warning: warning
        error: error
        critical: critical
    """

    debug = "debug"
    info = "info"
    warning = "warning"
    error = "error"
    critical = "critical"


class SysLog(Base):
    """System log table

    Fields:
        seq: sequence number
        time: time of log entry
        level: log level
        origin: origin of log entry
        message: log message
    """

    seq = Column(Integer, primary_key=True, index=True, autoincrement=True)
    time = Column(DateTime, default=func.now())
    level = Column(Integer)
    origin = Column(String)
    message = Column(String)


class SysLogSchema(BaseModel):
    """System log in the Timelink app

    Fields:
        seq: sequence number
        time: time of log entry
        level: log level
        origin: origin of log entry
        message: log message
    """

    seq: int
    time: datetime
    level: LogLevel
    origin: str
    message: str

    model_config = ConfigDict(from_attributes=True)


class SysLogCreateSchema(BaseModel):
    """System log create in the Timelink app

    Fields:
        level: log level
        origin: origin of log entry
        message: log message
    """

    level: LogLevel
    origin: str
    message: str

    model_config = ConfigDict(from_attributes=True)


class KleioImportedFile(Base):
    """Represents a Kleio file imported in the database

    Fields:
        path: path of the file
        name: name of the file
        structure: structure name
        translator: translator name
        translation_date: date of translation
        nerrors: number of errors
        nwarnings: number of warnings
        error_rpt: error report
        warning_rpt: warning report"""

    __tablename__ = "kleiofiles"

    """path: str"""
    path: Mapped[str] = mapped_column(String(1024), primary_key=True, nullable=False)
    """name: str"""
    name: Mapped[str] = mapped_column(String(255), nullable=False)
    structure: Mapped[str] = mapped_column(String(255), nullable=False)
    translator: Mapped[str] = mapped_column(String(255), nullable=False)
    translation_date: Mapped[datetime] = mapped_column(DateTime, nullable=False)
    nerrors: Mapped[int] = mapped_column(Integer, default=0, nullable=False)
    nwarnings: Mapped[int] = mapped_column(Integer, default=0, nullable=False)
    error_rpt: Mapped[Optional[str]] = mapped_column(String, nullable=True)
    warning_rpt: Mapped[Optional[str]] = mapped_column(String, nullable=True)
    imported: Mapped[datetime] = mapped_column(DateTime, nullable=False)
    imported_string: Mapped[str] = mapped_column(String(255), nullable=False)

    def __str__(self):
        return f"{self.name} ({self.path})"

    def __repr__(self):
        return (
            f"KleioImportedFile(path={self.path!r}, name={self.name!r}, "
            f"structure={self.structure!r}, translator={self.translator!r}, "
            f"translation_date={self.translation_date!r}, nerrors={self.nerrors!r}, "
            f"nwarnings={self.nwarnings!r}, error_rpt={self.error_rpt!r}, "
            f"warning_rpt={self.warning_rpt!r}, imported={self.imported!r}, "
            f"imported_string={self.imported_string!r})"
        )


class KleioImportedFileSchema(BaseModel):
    """Represents a Kleio file imported in the database

    Objects of this class contain the same fields as the KleioImportedFile class.
    """

    """path: str"""
    path: str
    """name: str"""
    name: str
    """structure: str structure used to translate the file"""
    structure: str
    """translator: str name of the translator"""
    translator: str
    """translation_date: datetime date of translation"""
    translation_date: datetime
    """nerrors: int number of errors in the translation"""
    nerrors: int
    """nwarnings: int number of warnings in the translation"""
    nwarnings: int
    """error_rpt: str error report"""
    error_rpt: Optional[str]
    """warning_rpt: str warning report"""
    warning_rpt: Optional[str]
    """imported: datetime date of import"""
    imported: datetime
    """imported_string: str date of import as string"""
    imported_string: str

    model_config = ConfigDict(from_attributes=True)



""" Utilities for styling pandas dataframes"""

from matplotlib import cm, colors


def category_palette(categories, cmap_name=None):
    """Create a color palette associated with a list of categories

    Args:
        Categories: List of categories
        cmap_name: matplotlib color map defaults to Pastel2
            see https://matplotlib.org/stable/tutorials/colors/colormaps.html

    Returns a dict with the categories as keys and colors as values
    """
    if cmap_name is None:
        cmap_name = "Pastel2"

    palette = cm.get_cmap(cmap_name, len(categories))
    mapcolors = [colors.rgb2hex(palette(i)) for i in range(len(categories))]
    cat_to_color = dict(zip(categories, mapcolors))
    return cat_to_color


def style_color_row_by_category(row, category="id", palette=None):
    """Color row by category. Function for styling dataframes
    Usage: display(df.style.apply(style_color_row_by_category,axis=1,palette=mypalette))
    Args
        row: this is passed by pandas when
        category:  column that determines the row color
        palette: a dict that maps category values to colors
    """
    id = row[category]
    row_colors = [f"background-color: {palette.get(id,'#ffffff')}"] * len(row)
    return row_colors


def styler_row_colors(df, category="id", columns=None, cmap_name="Pastel2"):
    """returns a dataframe setting the row color according to a category

    Args:
        df: dataframe
        category: name of column with category that determines color, defaults to id
        cmap_name; name of matplolib color map to use, defaults to 'Pastel2'

    """
    categories = df[category].unique()
    palette = category_palette(categories, cmap_name=cmap_name)
    if columns is None:
        columns = df.columns
    if category not in columns:
        cols = columns + [category]
    else:
        cols = columns

    r = df[cols].style.apply(
        style_color_row_by_category, axis=1, category=category, palette=palette
    )
    return r



"""Schemas for the Timelink API

In the FastAPi tutorial this file
is used for the pydantic models for the API,
including for the classes that are used for database access.
Another file, called models.py, is used in the tutorial
for the SQLAlchemy models.

We use a module called "models" for the SQLAlchemy models

Here we put the pydantic models that are not related to
database models like search requests and search results.

"""

# pylint: disable=too-few-public-methods

from typing import List, Optional
from datetime import datetime
from datetime import date
from pydantic import BaseModel, ConfigDict  # pylint: disable=import-error


class EntitySchema(BaseModel):
    """Pydantic Schema for Entity"""

    id: str
    pom_class: str
    inside: Optional[str]
    the_order: Optional[int]
    the_level: Optional[int]
    the_line: Optional[int]
    groupname: Optional[str]
    updated: Optional[datetime]
    indexed: Optional[datetime]
    contains: Optional[List["EntitySchema"]]

    model_config = ConfigDict(from_attributes=True)


class EntityBriefSchema(BaseModel):
    """Pydantic Schema for Entity brief

    No links to other entities
    """

    id: str
    pom_class: str
    inside: Optional[str]
    the_order: Optional[int]
    the_level: Optional[int]
    the_line: Optional[int]
    groupname: Optional[str]
    updated: Optional[datetime]
    indexed: Optional[datetime]

    model_config = ConfigDict(from_attributes=True)


class RelationSchema(BaseModel):
    id: str
    origin: str
    destination: str
    the_type: str
    the_value: str
    the_date: str
    obs: Optional[str]

    model_config = ConfigDict(from_attributes=True)


class RelationOutSchema(RelationSchema):
    dest_name: Optional[str]


class RelationInSchema(RelationSchema):
    org_name: Optional[str]


class AttributeSchema(BaseModel):
    entity: str
    the_type: str
    the_value: str
    the_date: str
    obs: Optional[str]
    groupname: str = None

    model_config = ConfigDict(from_attributes=True)


class EntityAttrRelSchema(BaseModel):
    """Pydantic Schema for Entity with attributes, relation and contained entities"""

    id: str
    pom_class: str
    inside: Optional[str]
    the_order: Optional[int]
    the_level: Optional[int]
    the_line: Optional[int]
    groupname: Optional[str]
    updated: Optional[datetime]
    indexed: Optional[datetime]
    attributes: Optional[List["AttributeSchema"]]
    rels_in: Optional[List["RelationInSchema"]]
    rels_out: Optional[List["RelationOutSchema"]]
    contains: Optional[List["EntityBriefSchema"]]

    model_config = ConfigDict(from_attributes=True)


class SearchRequest(BaseModel):
    """Search request

    Fields:
        q: search query
        after: date after which to search, possibly None
        until: date until which to search, possibly None
        skip: number of items to skip, default 0
        limit: number of items to return, default 100

    """

    q: str
    after: date | None = (
        None  # see https://docs.pydantic.dev/usage/types/#datetime-types
    )
    until: date | None = None
    skip: int | None = 0
    limit: int | None = 100


class SearchResults(BaseModel):
    """Search results

    Fields:
        results: list of search results
    """

    id: str
    the_class: str
    description: str
    start_date: date
    end_date: date


class ImportStats(BaseModel):
    """Import statistics

    Fields:
        datetime: date and time of import
        machine: machine where import was done
        database: specific database where import was done
        file: file that was imported
        import_time_seconds: time in seconds that import took
        entities_processed: number of entities processed
        entity_rate: number of entities processed per second
        person_rate: number of persons processed per second
        nerrors: number of errors
        errors: list of errors
    """

    datetime: date
    machine: str
    database: str
    file: str
    import_time_seconds: float
    entities_processed: int
    entity_rate: float
    person_rate: float
    nerrors: int
    errors: List[str]



"""
(c) Joaquim Carvalho 2021.
MIT License, no warranties.
"""

# pylint: disable=import-error


from sqlalchemy import Column, String, ForeignKey

from timelink.kleio.utilities import quote_long_text
from timelink.mhk.models.entity import Entity


class Object(Entity):
    __tablename__ = "objects"

    id = Column(String, ForeignKey("entities.id"), primary_key=True)
    name = Column(String, index=True)
    the_type = Column(String(32), index=True)
    obs = Column(String)

    __mapper_args__ = {"polymorphic_identity": "object"}

    def __repr__(self):
        sr = super().__repr__()
        return (
            f"Object(id={sr}, "
            f'name="{self.name}", '
            f'the_type="{self.the_type}", '
            f"obs={self.obs}"
            f")"
        )

    def __str__(self):
        if self.name is None:
            name = ""
        else:
            name = self.name + "/"
        r = f"{self.groupname}${name}{quote_long_text(self.the_type)}/id={self.id}"
        if self.obs is not None:
            r = f"{r}  /obs={quote_long_text(self.obs)}"
        return r



# pylint: disable=import-error
# pylint: disable=no-member

from warnings import warn
from sqlalchemy import (
    MetaData,
    create_engine,
    select,
    inspect,
    engine,
)
from sqlalchemy.orm import sessionmaker  # pylint: disable=import-error

from timelink.mhk.models.base_class import Base
from timelink.mhk.models.pom_som_mapper import PomSomMapper
from timelink.mhk.models.base_mappings import pom_som_base_mappings

SQLALCHEMY_ECHO = False


class TimelinkMHK:
    """
    Provide access to a Timelink-MHK database

    Example:

        from timelink.mhk.model.db import TimelinkMHK

        dbsys = TimelinkMHK("sql alchemy connection string")

    """

    sa_engine: engine = None
    db_name: str = None
    conn_string = None
    metadata: MetaData = None

    def __init__(self, conn_string=None, sql_echo=False):
        """
        For the interaction of "engine" and "session" alternative
        configurations see
        https://docs.sqlalchemy.org/en/14/orm/session_basics.html

        :param conn_string: a sqlalchemy connection URL, str.
        """

        self.init_db(conn_string, sql_echo)

    def init_db(self, conn_string, sql_echo=False):
        """
        Create an engine from a connection string.
        Check if database is properly setup up
        by creating core tables, PomSomMappers
        and ORM classes. Database should be ready
        for import after this method is called.

        :param conn_string: SQLAlchemy connection string
        :return:
        """

        if conn_string is not None:
            self.sa_engine = create_engine(
                conn_string or self.conn_string, future=True, echo=sql_echo
            )
            self.conn_string = conn_string
        if self.conn_string is None:
            raise ValueError("No connection string available")

        self.session = sessionmaker(
            autocommit=False, autoflush=False, bind=self.sa_engine
        )

        with self.session(bind=self.sa_engine) as session:
            self.create_tables()
            session.commit()
            session.rollback()
            self.load_database_classes(session)
            self.ensure_all_mappings(session)
            session.commit()

    def get_engine(self) -> sa_engine:
        """Return sqlalchemy engine"""
        return self.sa_engine

    def get_session(self) -> sessionmaker:
        """Return sqlalchemy session"""
        return self.session

    def sa_engine(self) -> sa_engine:
        """DEPRECATED use TimelinkDB.get_engine()"""
        warn(
            "This method is deprecated, use get_engine()",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.get_engine()

    def get_metadata(self) -> MetaData:
        """Return sqlalchemy metada"""
        return self.metadata

    def create_tables(self):
        """
        Creates the tables from the current ORM metadata if needed

        :return: None
        """
        self.metadata = Base.metadata  # pylint: disable=ignore-no-member
        self.metadata.create_all(self.sa_engine)  # only creates if missing

    def load_database_classes(self, session):
        """
        Populates database with core Database classes
        :param session:
        :return:
        """

        # Check if the core tables are there
        existing_tables = self.table_names()
        base_tables = [v[0].table_name for v in pom_som_base_mappings.values()]
        missing = set(base_tables) - set(existing_tables)
        if len(missing) > 0:
            self.create_tables()

        # check if we have the data for the core database entity classes
        stmt = select(PomSomMapper.id)
        available_mappings = session.execute(stmt).scalars().all()
        for (
            k
        ) in (
            pom_som_base_mappings.keys()
        ):  # pylint: disable=consider-iterating-dictionary, consider-using-dict-items
            if k not in available_mappings:  # pylint: disable=consider-using-dict-items
                data = pom_som_base_mappings[k]
                session.bulk_save_objects(data)
        # this will cache the pomsom mapper objects
        session.commit()

    def ensure_all_mappings(self, session):
        """Ensure that all database classes have a table and ORM class"""
        pom_classes = PomSomMapper.get_pom_classes(session)
        for pom_class in pom_classes:
            pom_class.ensure_mapping(session)

    def table_names(self):
        """Current tables in the current database"""
        insp = inspect(self.sa_engine)
        db_tables = insp.get_table_names()  # tables in the database
        return db_tables

    def drop_db(self, session):
        """
        This will drop all timelink related tables from the database.
        It will not touch non timelink tables that might exist.
        :param session:
        :return:
        """
        session.rollback()
        self.load_database_classes(session)
        self.ensure_all_mappings(session)
        self.metadata = Base.metadata  # pylint: disable=ignore-no-member
        self.metadata.drop_all(self.sa_engine)


# for compatibility with old code
TimelinkDB = TimelinkMHK



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = ['typer~=0.4.0', ]

test_requirements = ['pytest>=3', ]

description = "Timelink is an information system for person related information collected from historical sources. "  # noqa

setup(
    author="Joaquim Ramos de Carvalho",
    author_email='joaquimcarvalho@ipm.edu.mo',
    python_requires='>=3.10',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],
    description=description,
    entry_points={
        'console_scripts': [
            'timelink=timelink.cli:app',
        ],
    },
    project_urls={
        'Documentation': 'https://timelink-py.readthedocs.io/',
    },
    install_requires=[
        'typer',
        'fastapi',
        'fastui',
        'fief-client',
        'sqlalchemy',
        'bcrypt==4.0.1',  # bug see https://github.com/logspace-ai/langflow/issues/1173
        'passlib[bcrypt]',
        'pydantic',
        'pydantic-settings',
        'python-dotenv',
        'python-box',
        'python-multipart',
        'psycopg2-binary',
        'py-markdown-table',
        'python-jose[cryptography]',
        'pandas',
        'docker',
        'jsonrpcclient',
        'httpx',
        'ipython',
        'uvicorn'
    ],
    license="MIT license",
    long_description=description + '\n\n' + history,
    long_description_content_type='text/x-rst',
    include_package_data=True,
    keywords='timelink',
    name='timelink',
    packages=find_packages(include=['timelink', 'timelink.*']),
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/time-link/timelink-py',
    version='1.1.5',
    zip_safe=False,
)



from timelink.kleio.groups.kgroup import KGroup
from timelink.kleio.groups.kact import KAct


class KObject(KGroup):
    """KObject(name,type,id=,obs=,same_as=,xsame_as=)

    An object in a historical source.
    Object groups represent physical entities like
    houses, pieces of land, movable objects

    Elements:
        name: the name of the object. A string.
        type:  the . A string.
        id: an unique id for this person. A string, optional.
        obs: a note on the person. A string, optional.
        same_as: id of another instance of this object in the same file.
        xsame_as: id of another instance of this object in another file.

    Kleio str definition:

        part name=object;
             guaranteed=name;
             position=name,type;
             also=obs,id,same_as,xsame_as;
             arbitrary=atr,ls,rel
    """

    _name = "object"
    _guaranteed = ["name"]
    _also = ["id", "obs", "same_as"]
    _position = ["name", "sex", "id", "same_as", "xsame_as"]
    _part = ["rel", "attr"]
    _pom_class_id: str = "object"


KAct.allow_as_part(KObject)



import logging
import os
import sys
import time
import copy

from datetime import datetime, timezone
from enum import Enum
from pathlib import Path
from typing import List, Optional, Union
import platform

import urllib.request

from xml.sax import handler, make_parser, SAXParseException

from sqlalchemy import func, delete  # pylint: disable=import-error
from sqlalchemy.orm import Session  # pylint: disable=import-error
from sqlalchemy.exc import IntegrityError  # pylint: disable=import-error

from timelink.kleio.groups import KGroup, KElement
from timelink.mhk.models.db import pom_som_base_mappings as pom_som_base_mappingsMHK
from timelink.mhk.models.pom_som_mapper import (
    PomSomMapper as PomSomMapperMHK,
    PomClassAttributes as PomClassAttributesMHK,
)
from timelink.mhk.models.entity import Entity as EntityMHK
from timelink.mhk.models.person import Person as PersonMHK
from timelink.mhk.models.system import KleioImportedFile as KleioFileMHK

from timelink.api.models.base_mappings import (
    pom_som_base_mappings as pom_som_base_mappingsTL,
)
from timelink.api.models.pom_som_mapper import PomSomMapper as PomSomMapperTL
from timelink.api.models.pom_som_mapper import (
    PomClassAttributes as PomClassAttributesTL,
)
from timelink.api.models.base import Entity as EntityTL, Person as PersonTL
from timelink.api.models.system import KleioImportedFile as KleioFileTL


class KleioContext(Enum):
    """Kleio context enumeration"""

    START = (1,)
    KLEIO = (2,)
    CLASS = (3,)
    GROUP = (4,)
    ELEMENT = (5,)
    CORE = (6,)
    COMMENT = 7
    ORIGINAL = 8


class SaxHandler(handler.ContentHandler):
    """SAX Handler for Kleio XML files"""

    _context: KleioContext
    _current_class: Optional[PomSomMapperTL | PomSomMapperMHK]
    _current_class_attrs: List[PomClassAttributesTL | PomClassAttributesMHK]
    _current_group: Optional[KGroup]
    _current_element: Optional[KElement]
    _current_entry: str

    def __init__(self, kleio_handler):
        handler.ContentHandler.__init__(self)
        self._kleio_handler = kleio_handler

    def startDocument(self):
        self._context = KleioContext.START
        self._current_class = None
        self._current_class_attrs = []
        self._current_group = None
        self._current_element = None
        self._current_entry = ""

    def endDocument(self):
        self._kleio_handler.endKleioFile()

    def startElement(self, name: str, attrs):
        #  https://stackoverflow.com/questions/15477363/xml-sax-parser-and-line-numbers-etc
        loc = self._locator
        ename = name.upper()
        if ename == "KLEIO":
            self._kleio_handler.newKleioFile(attrs)
            self._context = KleioContext.KLEIO
        elif ename == "CLASS":
            # <CLASS NAME="source" SUPER="entity"
            #       TABLE="sources" GROUP="fonte">
            if self._context != KleioContext.KLEIO:
                raise SAXParseException(
                    exception=None,
                    locator=loc,
                    msg="CLASS out of context, should be inside KLEIO element",
                )
            self._current_class = self._kleio_handler.pom_som_mapper(
                id=attrs["NAME"],
                super_class=attrs["SUPER"],
                table_name=attrs["TABLE"],
                group_name=attrs["GROUP"],
            )
            self._current_class_attrs = []
            self._context = KleioContext.CLASS

        elif ename == "ATTRIBUTE":
            #  <ATTRIBUTE NAME="id"
            #       COLUMN="id"
            #       CLASS="id"
            #       TYPE="varchar"
            #       SIZE="64"
            #       PRECISION="0"
            #       PKEY="1" >
            #   </ATTRIBUTE>
            if self._context != KleioContext.CLASS:
                raise SAXParseException(
                    "ATTRIBUTE out of context, should be inside CLASS element",
                    None,
                    loc,
                )
            new_attribute = self._kleio_handler.pom_class_attributes()
            new_attribute.the_class = self._current_class.id
            new_attribute.name = attrs["NAME"]
            new_attribute.colname = attrs["COLUMN"]
            new_attribute.colclass = attrs["CLASS"]
            new_attribute.coltype = attrs["TYPE"]
            new_attribute.colsize = int(attrs["SIZE"])
            new_attribute.colprecision = int(attrs["PRECISION"])
            new_attribute.pkey = int(attrs["PKEY"])

            self._current_class_attrs.append(new_attribute)

        elif ename == "GROUP":
            if self._context != KleioContext.KLEIO:
                raise SAXParseException(
                    "GROUP out of context, should be inside KLEIO element", None, loc
                )
            # <GROUP ID="coja-rol-1841" NAME="fonte"
            #        CLASS="source" ORDER="1" LEVEL="1" LINE="4">
            id = attrs["ID"]
            name = attrs["NAME"]
            the_class = attrs["CLASS"]
            order = attrs["ORDER"]
            level = attrs["LEVEL"]
            line = attrs["LINE"]

            # Mal [? porquê?]
            self._current_group = KGroup()
            self._current_group.id = id
            self._current_group._name = name
            self._current_group._pom_class_id = the_class
            self._current_group._order = int(order)
            self._current_group.level = int(level)
            self._current_group.line = int(line)
            self._context = KleioContext.GROUP

        elif ename == "ELEMENT":
            if self._context != KleioContext.GROUP:
                raise SAXParseException(
                    "ELEMENT out of context, should be inside GROUP element",
                    exception=None,
                    locator=loc,
                )
            # <ELEMENT NAME="id" CLASS="id">
            #       <core>r1775-f1-per1</core></ELEMENT>

            elname = attrs["NAME"]
            self._current_group.allow_as_element(elname)
            self._current_element = KElement(elname, None)
            self._current_element.element_class = attrs["CLASS"]
            self._context = KleioContext.ELEMENT

        elif ename == "CORE":
            if self._context != KleioContext.ELEMENT:
                raise SAXParseException(
                    "CORE out of context, should be inside ELEMENT element",
                    exception=None,
                    locator=loc,
                )
            self._context = KleioContext.CORE
            self._current_entry = ""

        elif ename == "COMMENT":
            if self._context != KleioContext.ELEMENT:
                raise SAXParseException(
                    "COMMENT out of context, should be inside ELEMENT element",
                    exception=None,
                    locator=loc,
                )
            self._context = KleioContext.COMMENT
            self._current_entry = ""

        elif ename == "ORIGINAL":
            if self._context != KleioContext.ELEMENT:
                raise SAXParseException(
                    "ORIGINAL out of context, should be inside ELEMENT element",
                    exception=None,
                    locator=loc,
                )
            self._context = KleioContext.ORIGINAL
            self._current_entry = ""

        elif ename == "RELATION":
            if self._context != KleioContext.KLEIO:
                raise SAXParseException(
                    "GROUP out of context, should be inside KLEIO element",
                    exception=None,
                    locator=loc,
                )
            self._kleio_handler.newRelation(attrs)
        else:
            raise SAXParseException(
                f"Unexpected element in Kleio file: {ename}",
                exception=None,
                locator=loc,
            )

    def endElement(self, name):
        ename = name.upper()

        if ename == "GROUP":
            self._kleio_handler.newGroup(self._current_group)
            self._context = KleioContext.KLEIO
            self._current_group = None

        elif ename == "CLASS":
            self._kleio_handler.newClass(self._current_class, self._current_class_attrs)
            self._context = KleioContext.KLEIO

        elif ename == "ATTRIBUTE":
            self._context = KleioContext.CLASS

        elif ename == "ELEMENT":
            self._current_group[self._current_element.name] = self._current_element

            self._context = KleioContext.GROUP

        elif ename in ["CORE", "COMMENT", "ORIGINAL"]:
            if ename == "CORE":
                if self._current_element.core is None:
                    self._current_element.core = self._current_entry
            if ename == "COMMENT":
                if self._current_element.comment is None:
                    self._current_element.comment = self._current_entry
            if ename == "ORIGINAL":
                if self._current_element.original is None:
                    self._current_element.original = self._current_entry
            self._context = KleioContext.ELEMENT

    def characters(self, content):
        self._current_entry = self._current_entry + content

    def ignorableWhitespace(self, whitespace):
        pass

    def processingInstruction(self, target, data):
        pass


class KleioHandler:
    pom_som_base_mappings = None
    pom_som_mapper = None
    pom_class_attributes = None
    entity_model = None
    person_model = None
    kleio_file_model = None
    model_type: str = None
    postponed_relations = []
    errors = []
    warnings = []
    kleio_file_origin = None
    kleio_file = None
    kleio_file_name = None
    kleio_structure = None
    kleio_translator = None
    kleio_when = None
    kleio_context = None
    pom_som_cache = dict()

    def __init__(self, session, mode="TL"):
        """
        Arguments:
            session: a SQLAlchemy session
            mode: the mode of the import, either 'TL'(Timelink) or 'MHK'

        """
        self.session = session
        if mode == "TL":
            self.pom_som_base_mappings = pom_som_base_mappingsTL
            self.pom_som_mapper = PomSomMapperTL
            self.pom_class_attributes = PomClassAttributesTL
            self.pom_class_attributes = PomClassAttributesTL
            self.entity_model = EntityTL
            self.person_model = PersonTL
            self.kleio_file_model = KleioFileTL
            self.model_type = "TL"
        elif mode == "MHK":
            self.pom_som_base_mappings = pom_som_base_mappingsMHK
            self.pom_som_mapper = PomSomMapperMHK
            self.pom_class_attributes = PomClassAttributesMHK
            self.entity_model = EntityMHK
            self.person_model = PersonMHK
            self.kleio_file_model = KleioFileMHK
            self.model_type = "MHK"
        else:
            raise ValueError(f"Unknown mode {mode}")

    def newKleioFile(self, attrs):
        # <KLEIO STRUCTURE="/kleio-home/system/conf/kleio/stru/gacto2.str"
        # SOURCE="/kleio-home/sources/soure-fontes/sources/1685-1720/baptismos/b1685.cli"
        # TRANSLATOR="gactoxml2.str"
        # WHEN="2020-8-18 17:10:32"
        # OBS="" SPACE="">

        self.session.commit()
        self.postponed_relations = []
        self.errors = []
        self.warnings = []
        self.kleio_file = attrs["SOURCE"]
        # extract name of file from path
        self.kleio_file_name = self.kleio_file.split("/")[-1]
        self.kleio_structure = attrs["STRUCTURE"]
        self.kleio_translator = attrs["TRANSLATOR"]
        self.kleio_when = attrs["WHEN"]
        self.pom_som_cache = dict()
        logging.debug("Kleio file: %s", self.kleio_file)
        logging.debug("Kleio structure: %s", self.kleio_structure)
        logging.debug("Kleio translator: %s", self.kleio_translator)
        logging.debug("Kleio when: %s", self.kleio_when)

    def newClass(
        self,
        psm: PomSomMapperTL | PomSomMapperMHK,
        attrs: List[PomClassAttributesTL | PomClassAttributesMHK],
    ):
        if psm.id in self.pom_som_base_mappings.keys():
            # we do not allow redefining of base mappings
            logging.debug("Skipping base mapping %s", psm.id)
            return

        session = self.session

        # check if we have this class defined in the database.
        existing_psm: PomSomMapperTL | PomSomMapperMHK = session.get(
            self.pom_som_mapper, psm.id
        )
        if existing_psm is not None:  # class exists: delete, insert again
            try:
                logging.debug("Deleting existing class %s", psm.id)
                session.commit()
                stmt = delete(self.pom_class_attributes).where(
                    self.pom_class_attributes.the_class == psm.id
                )
                session.execute(stmt)
                session.delete(existing_psm)
                session.commit()
            except Exception as e:
                session.rollback()
                self.errors.append(
                    f"ERROR: deleting class {psm.id}: {e.__class__.__name__}: {e}"
                )

        # now we add the new mapping
        try:
            logging.debug("Adding class %s", psm.id)
            session.add(psm)
            for class_attr in attrs:
                session.add(class_attr)
            session.commit()
        except Exception as e:
            session.rollback()
            self.errors.append(
                f"ERROR: adding class {psm.id}: {e.__class__.__name__}: {e}"
            )
        # ensure that the table and ORM classes are created
        try:
            psm.ensure_mapping(session)
            session.commit()
        except Exception as e:
            session.rollback()
            self.errors.append(
                f"ERROR: creating ORM mapping for class {psm.id}: {e.__class__.__name__}: {e}"
            )

    def newGroup(self, group: KGroup):
        # get the PomSomMapper
        # pass storeKGroup
        pom_mapper_for_group: PomSomMapperTL | PomSomMapperMHK
        try:
            if group.pom_class_id in self.pom_som_cache.keys():
                pom_mapper_for_group = self.pom_som_cache[group.pom_class_id]
            else:
                pom_mapper_for_group = self.pom_som_mapper.get_pom_class(
                    group.pom_class_id, self.session
                )
            if pom_mapper_for_group is None:
                raise ValueError(
                    f"Could not find PomSomMapper for class {group.pom_class_id}"
                )
            self.pom_som_cache[group.pom_class_id] = pom_mapper_for_group
        except Exception as exc:
            self.errors.append(
                f"ERROR: {self.kleio_file_name} {str(group.line)}"
                f" finding PomSomMapper for class "
                f"{group.pom_class_id}: {exc.__class__.__name__}: {exc}"
            )
            return

        if group.pom_class_id == "act":
            logging.debug(f"{group._name}${group.id}")

        if pom_mapper_for_group.id == "relation":
            # it can happen that the destination of a relation
            #  is not yet in the database (forward reference in relation)
            #  In this case we postpone the storing of the relation
            # until the end of file
            exits_dest_rel = self.session.get(
                self.entity_model, group.get_element_for_column("destination").core
            )
            if exits_dest_rel is None:
                self.postponed_relations.append(
                    (pom_mapper_for_group.id, copy.deepcopy(group))
                )
            else:
                try:
                    pom_mapper_for_group.store_KGroup(group, self.session)
                except Exception as exc:
                    self.session.rollback()
                    self.errors.append(
                        f"ERROR: {self.kleio_file_name} {str(group.line)} "
                        f"storing group {group.kname}${group.id}: "
                        f"{exc.__class__.__name__}: {exc}"
                    )
                    self.session.rollback()
        else:
            try:
                pom_mapper_for_group.store_KGroup(group, self.session)
            except IntegrityError as ierror:
                self.errors.append(
                    f"ERROR: {self.kleio_file_name} line {str(group.line)} "
                    f"** integrity error {group.kname}${group.id}: {ierror}"
                )
                self.session.rollback()

            except Exception as exc:
                self.session.rollback()
                self.errors.append(
                    f"ERROR: {self.kleio_file_name} line {str(group.line)} "
                    f"storing group {group.kname}${group.id}: {exc.__class__.__name__}: {exc}"
                )

    def newRelation(self, attrs):
        pass

    def endKleioFile(self):
        """Process end of file: process postponed relations"""
        # store postponed relations
        postponed = len(self.postponed_relations)
        if postponed > 0:
            log = f"Storing {postponed} postponed relations"
            logging.info(log)
        for pom_mapper_id, group in self.postponed_relations:
            pom_mapper = self.pom_som_mapper.get_pom_class(pom_mapper_id, self.session)
            try:
                pom_mapper.store_KGroup(group, self.session)
                self.session.commit()
            except IntegrityError as ierror:
                self.errors.append(
                    f"ERROR: {self.kleio_file_name} {str(group.line)} "
                    f"storing {group.to_kleio()}\n {ierror}"
                )
                self.session.rollback()
            except Exception as exc:
                self.errors.append(
                    f"ERROR: {self.kleio_file_name} {str(group.line)} "
                    f"storing {group.to_kleio()}\n {exc.__class__.__name__}: {exc}"
                )
                self.session.rollback()
        self.postponed_relations = []
        # store information on kleio file import
        kfile = self.kleio_file_model(
            path=self.kleio_file,  # TODO: #20 should remove /Kleio-home from path
            name=self.kleio_file_name,
            structure=self.kleio_structure,
            translator=self.kleio_translator,
            # convert date from "2020-8-18 17:10:32" format
            # to datetime object
            translation_date=datetime.strptime(self.kleio_when, "%Y-%m-%d %H:%M:%S"),
            nerrors=len(self.errors),
            nwarnings=len(self.warnings),
        )
        if len(self.errors) > 0:
            s = "\n\n".join(self.errors)
        else:
            s = "No errors"
        kfile.error_rpt = s
        if len(self.warnings) > 0:
            s = "\n\n".join(self.warnings)
        else:
            s = "No warnings"
        kfile.warning_rpt = s
        kfile.imported = datetime.now(timezone.utc)
        kfile.imported_string = datetime.now(timezone.utc).strftime(
            "%Y-%m-%d %H:%M:%S %Z"
        )
        kfile_exists = self.session.get(self.kleio_file_model, kfile.path)
        if kfile_exists is not None:
            self.session.delete(kfile_exists)
            self.session.commit()
        self.session.add(kfile)
        self.session.commit()


def import_from_xml(
    filespec: Union[str, Path], session: Session, options: dict = None
) -> dict:
    """Import data from file or url into a timelink-mhk database.

    The data file must be a XML file in the format exported by the ``kleio``
    translator.

    Arguments:
        filespec (str,Path): a file path, URL of a data file or path in a Kleio Server.
        session: a database session from TimelinkDatabase.session()
        options (dict): a dictionnary with options

           - 'return_stats':  if True import stats will be returned
           - 'kleio_url':  the url of kleio server;
           - 'kleio_token':  the authorization token for the kleio server.
           - 'mode':  the mode of the import, either 'TL'(Timelink) or 'MHK'

        If kleio_url and kleio_token are specified the data will be fetched from
        a KleioServer and the filespec should contain the "xml_path" of the file
        in the server (e.g. /rest/exports/reference_sources/varia/vereacao.xml) and
        the kleio_token will be inserted in the header for autentication

    Returns:
        If stats is True in options a dict with statistical information will be
        returned.

            - 'datetime': the time stamp of the start of the import
            - 'machine':  local machine name
            - 'file': path to imported file or url
            - 'import_time_seconds':  elapsed time during import
            - 'entities_processed': number of entities processed
            - 'entity_rate': number of entities processed per second
            - 'person_rate': number of persons (entities of class 'person')
            - 'nerrors': number of errors during import
            - 'errors': list of error messages

    Examples:
        Returned statistical information when stats=True
    ::

        {
        'datetime': '2022-01-02 18:11:12',
        'machine': 'joaquims-mbpr.local',
        'file': 'https://...b1685.xml',
        'import_time_seconds': 7.022288084030151,
        'entities_processed': 747,
        'entity_rate': 106.37558457603042,
        'person_rate': 27.483919441999827
        'nerrors': 0
        'errors': []
        }

        TODO: should use https when the kleio_url not local.
    """
    collect_stats = False
    kleio_url = None
    kleio_token = None
    mode = "TL"  # determines the database model TL=Timelink, MHK=MHK
    nentities_before = 0
    npersons_before = 0
    now = datetime.now()
    if options is not None and options.get("return_stats", False):
        collect_stats = True
    if options is not None and options.get("mode", None) is not None:
        mode = options.get("mode")
    if options is not None and options.get("kleio_url", None) is not None:
        kleio_url = options.get("kleio_url")
    if options is not None and options.get("kleio_token", None) is not None:
        kleio_token = options.get("kleio_token")

    kleio_handler = KleioHandler(session, mode=mode)
    sax_handler = SaxHandler(kleio_handler)
    parser = make_parser()
    parser.setContentHandler(sax_handler)
    start = time.time()
    if collect_stats:
        nentities_before = session.query(
            func.count(kleio_handler.entity_model.id)
        ).scalar()
        npersons_before = session.query(
            func.count(kleio_handler.person_model.id)
        ).scalar()

    if kleio_url is not None and kleio_token is not None:
        headers = {"Authorization": f"Bearer {kleio_token}"}
        server_url = f"{kleio_url}{filespec}"
        req = urllib.request.Request(server_url, headers=headers)
        with urllib.request.urlopen(req, timeout=30) as source:
            parser.parse(source)
    elif kleio_token is not None or kleio_url is not None:
        # this means that one of the options is missing
        raise ValueError(
            "Both kleio_url and kleio_token must be specified to fetch from kleio server"
        )
    elif isinstance(filespec, os.PathLike):
        source = os.fspath(filespec)
        parser.parse(source)
    else:
        source = filespec
        parser.parse(source)

    end = time.time()
    if collect_stats:
        machine = platform.node()
        nentities = session.query(func.count(kleio_handler.entity_model.id)).scalar()
        npersons = session.query(func.count(kleio_handler.person_model.id)).scalar()
        erate = (nentities - nentities_before) / (end - start)
        prate = (npersons - npersons_before) / (end - start)
        stats = {
            "datetime": now.timestamp(),
            "machine": machine,
            "database": session.bind.url,
            "file": filespec,
            "import_time_seconds": end - start,
            "entities_processed": nentities - nentities_before,
            "entity_rate": erate,
            "person_rate": prate,
            "nerrors": len(kleio_handler.errors),
            "errors": kleio_handler.errors,
        }
        return stats


if (
    __name__ == "__main__"
):  # Not sure this works, where is the session for KleioHandler?
    saxParser = make_parser()
    kleioHandler = KleioHandler(None, mode="TL")
    saxParser.setContentHandler(SaxHandler(kleioHandler))
    saxParser.parse(sys.argv[1])



# pylint: disable=import-error
from sqlalchemy import Column, String, ForeignKey, Index
from sqlalchemy.orm import relationship

from timelink.kleio.utilities import kleio_escape, quote_long_text
from .entity import Entity


class Relation(Entity):
    __tablename__ = "relations"

    id = Column(String, ForeignKey("entities.id", ondelete="CASCADE"), primary_key=True)
    # rel_entity = relationship("Entity",
    #                   foreign_keys='id',back_populates='rel')
    origin = Column(String, ForeignKey("entities.id"), index=True)
    org = relationship(Entity, foreign_keys=[origin], back_populates="rels_out")

    destination = Column(String, ForeignKey("entities.id"), index=True)
    dest = relationship("Entity", foreign_keys=[destination], back_populates="rels_in")
    the_type = Column(String, index=True)
    the_value = Column(String, index=True)
    the_date = Column(String, index=True)
    obs = Column(String)

    __mapper_args__ = {
        "polymorphic_identity": "relation",
        "inherit_condition": id == Entity.id,
    }

    __table_args__ = (Index("relations_type_value", "the_type", "the_value"),)

    @property
    def dest_class(self):
        return self.dest.pom_class

    @property
    def org_class(self):
        return self.org.pom_class

    @property
    def dest_name(self):
        if self.dest_class == "person" or self.dest_class == "object":
            return self.dest.name
        else:
            return self.dest.groupname

    @property
    def org_name(self):
        if self.org_class == "person" or self.org_class == "object":
            return self.org.name
        else:
            return self.org.groupname

    def __repr__(self):
        sr = super().__repr__()
        return (
            f"Relation(id={sr}, "
            f'origin="{self.origin}", '
            f'destination="{self.destination}", '
            f'the_type="{self.the_type}", '
            f'the_value="{self.the_value}", '
            f'the_date="{self.the_date}"", '
            f"obs={self.obs}"
            f")"
        )

    def __str__(self):
        if self.dest is not None and self.dest.pom_class == "person":
            r = (
                f"rel${kleio_escape(self.the_type)}/{quote_long_text(self.the_value)}/{kleio_escape(self.dest.name)}"
                f"/{self.destination}/{self.the_date}"
            )
        else:
            r = (
                f"rel${self.the_type}/{quote_long_text(self.the_value)}/"
                f"{self.destination}/{self.the_date}"
            )
        if self.obs is not None:
            r = f"{r}/obs={quote_long_text(self.obs)}"
        return r

    def to_kleio(self, ident="", ident_inc="  "):
        if self.the_type == "function-in-act":
            return ""
        else:
            # call to_kleio from the parent class
            return super().to_kleio(ident, ident_inc)


Entity.rels_out = relationship(
    "Relation", foreign_keys=[Relation.origin], back_populates="dest"
)

Entity.rels_in = relationship(
    "Relation", foreign_keys=[Relation.destination], back_populates="org"
)



from timelink.kleio.groups.kgroup import KGroup
from timelink.kleio.groups.ksource import KSource


class KAct(KGroup):
    """KAct(id,type,date[,day=,month=,year=,loc=,ref=,obs=])

    An Act is a record of an event in a specific date.

    Fields:

        id: an unique id for this act. A string.
        type: type of the act (baptism, marriage, contract...). A string.
        date: the date of the act. A string in timelink format.
        day,month,year: the date expressed in individual values
        loc: location of the act in the archive (if different from source)
        ref: call number, or page number in source.

    Kleio str definition:

    part 	name=historical-act;
            guaranteed=id,type,date;
            position=id,type,date;
            also=loc,ref,obs,day,month,year;
            arbitrary=person,object,geoentity,abstraction,ls,atr,rel

    """

    _name = "act"
    _guaranteed = ["id", "type", "date"]
    _position = ["id", "type", "date"]
    _also = ["loc", "ref", "obs", "day", "month", "year"]
    _part = ["person", "object", "geoentity", "abstraction", "ls", "attr", "rel"]
    _pom_class_id: str = "act"


KSource.allow_as_part(KAct)
